# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""d4rl_adroit_relocate dataset."""

from tensorflow_datasets.d4rl.d4rl_adroit_relocate import d4rl_adroit_relocate
import tensorflow_datasets.public_api as tfds


class D4rlAdroitRelocateClonedV1Test(tfds.testing.DatasetBuilderTestCase):
  """Tests for d4rl_adroit_relocate dataset."""

  DATASET_CLASS = d4rl_adroit_relocate.D4rlAdroitRelocate

  SPLITS = {
      'train': 2,  # Number of fake train example
  }
  SKIP_TF1_GRAPH_MODE = True

  DL_EXTRACT_RESULT = {'file_path': 'relocate-cloned-v1.hdf5'}
  DL_DOWNLOAD_RESULT = {'file_path': 'relocate-cloned-v1.hdf5'}

  BUILDER_CONFIG_NAMES_TO_TEST = ['v1-cloned']


class D4rlAdroitRelocateExpertV1Test(tfds.testing.DatasetBuilderTestCase):
  """Tests for d4rl_adroit_relocate dataset."""

  DATASET_CLASS = d4rl_adroit_relocate.D4rlAdroitRelocate

  SPLITS = {
      'train': 2,  # Number of fake train example
  }
  SKIP_TF1_GRAPH_MODE = True

  DL_EXTRACT_RESULT = {'file_path': 'relocate-expert-v1.hdf5'}
  DL_DOWNLOAD_RESULT = {'file_path': 'relocate-expert-v1.hdf5'}

  BUILDER_CONFIG_NAMES_TO_TEST = ['v1-expert']


class D4rlAdroitRelocateHumanV1Test(tfds.testing.DatasetBuilderTestCase):
  """Tests for d4rl_adroit_relocate dataset."""

  DATASET_CLASS = d4rl_adroit_relocate.D4rlAdroitRelocate

  SPLITS = {
      'train': 2,  # Number of fake train example
  }
  SKIP_TF1_GRAPH_MODE = True

  DL_EXTRACT_RESULT = {'file_path': 'relocate-human-v1.hdf5'}
  DL_DOWNLOAD_RESULT = {'file_path': 'relocate-human-v1.hdf5'}

  BUILDER_CONFIG_NAMES_TO_TEST = ['v1-human']


class D4rlAdroitRelocateClonedV0Test(tfds.testing.DatasetBuilderTestCase):
  """Tests for d4rl_adroit_relocate dataset."""

  DATASET_CLASS = d4rl_adroit_relocate.D4rlAdroitRelocate

  SPLITS = {
      'train': 2,  # Number of fake train example
  }
  SKIP_TF1_GRAPH_MODE = True

  DL_EXTRACT_RESULT = {'file_path': 'relocate-cloned-v0.hdf5'}
  DL_DOWNLOAD_RESULT = {'file_path': 'relocate-cloned-v0.hdf5'}

  BUILDER_CONFIG_NAMES_TO_TEST = ['v0-cloned']


class D4rlAdroitRelocateExpertV0Test(tfds.testing.DatasetBuilderTestCase):
  """Tests for d4rl_adroit_relocate dataset."""

  DATASET_CLASS = d4rl_adroit_relocate.D4rlAdroitRelocate

  SPLITS = {
      'train': 2,  # Number of fake train example
  }
  SKIP_TF1_GRAPH_MODE = True

  DL_EXTRACT_RESULT = {'file_path': 'relocate-expert-v0.hdf5'}
  DL_DOWNLOAD_RESULT = {'file_path': 'relocate-expert-v0.hdf5'}

  BUILDER_CONFIG_NAMES_TO_TEST = ['v0-expert']


class D4rlAdroitRelocateHumanV0Test(tfds.testing.DatasetBuilderTestCase):
  """Tests for d4rl_adroit_relocate dataset."""

  DATASET_CLASS = d4rl_adroit_relocate.D4rlAdroitRelocate

  SPLITS = {
      'train': 2,  # Number of fake train example
  }
  SKIP_TF1_GRAPH_MODE = True

  DL_EXTRACT_RESULT = {'file_path': 'relocate-human-v0.hdf5'}
  DL_DOWNLOAD_RESULT = {'file_path': 'relocate-human-v0.hdf5'}

  BUILDER_CONFIG_NAMES_TO_TEST = ['v0-human']


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.example_parser."""

import numpy as np
import pytest
from tensorflow_datasets.core import example_parser
from tensorflow_datasets.core import features as features_lib
from tensorflow_datasets.proto import tf_example_pb2
from tensorflow_datasets.proto import tf_feature_pb2


def test_example_parser_np():
  features = features_lib.FeaturesDict({
      'feature': features_lib.FeaturesDict({
          'nested_text': features_lib.Text(),
          'nested_float': np.float32,
      }),
      'array_of_ints': features_lib.Tensor(shape=(2,), dtype=np.int8),
      'array_of_unknown_shape': features_lib.Tensor(
          shape=(None,), dtype=np.uint16
      ),
  })
  serialized_example = tf_example_pb2.Example(
      features=tf_feature_pb2.Features(
          feature={
              'feature/nested_text': tf_feature_pb2.Feature(
                  bytes_list=tf_feature_pb2.BytesList(value=[b'thisistext'])
              ),
              'feature/nested_float': tf_feature_pb2.Feature(
                  float_list=tf_feature_pb2.FloatList(value=[1.0])
              ),
              'array_of_ints': tf_feature_pb2.Feature(
                  int64_list=tf_feature_pb2.Int64List(value=[2, 3])
              ),
              'array_of_unknown_shape': tf_feature_pb2.Feature(
                  int64_list=tf_feature_pb2.Int64List(value=[4, 5, 6])
              ),
          }
      )
  ).SerializeToString()
  expected_np = {
      'feature': {'nested_text': b'thisistext', 'nested_float': 1.0},
      'array_of_ints': np.asarray([2, 3], dtype=np.int8),
      'array_of_unknown_shape': np.asarray([4, 5, 6], dtype=np.uint16),
  }
  example_specs = features.get_tensor_info()
  example_parser_np = example_parser.ExampleParserNp(example_specs)
  parsed_example = example_parser_np.parse_example(serialized_example)
  assert isinstance(parsed_example['array_of_ints'], np.ndarray)
  assert isinstance(parsed_example['array_of_unknown_shape'], np.ndarray)
  assert parsed_example['array_of_ints'].dtype == np.int8
  assert parsed_example['array_of_unknown_shape'].dtype == np.uint16
  np.testing.assert_equal(parsed_example, expected_np)


def test_key_error_exception_if_example_specs_is_malformed():
  features = features_lib.FeaturesDict({'doesnotexist': features_lib.Text()})
  serialized_example = tf_example_pb2.Example(
      features=tf_feature_pb2.Features(
          feature={
              'array_of_ints': tf_feature_pb2.Feature(
                  int64_list=tf_feature_pb2.Int64List(value=[2, 3, 4])
              ),
          }
      )
  ).SerializeToString()
  example_specs = features.get_tensor_info()
  example_parser_np = example_parser.ExampleParserNp(example_specs)
  with pytest.raises(
      KeyError,
      match='(.|\n)*array_of_ints is found in the feature, but not in*',
  ):
    example_parser_np.parse_example(serialized_example)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Grounded Scan dataset."""
from __future__ import annotations

import json
import os
from typing import List

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """
Grounded SCAN (gSCAN) is a synthetic dataset for evaluating compositional
generalization in situated language understanding. gSCAN pairs natural language
instructions with action sequences, and requires the agent to interpret
instructions within the context of a grid-based visual navigation environment.

More information can be found at:

* For the `compositional_splits` and the `target_length_split`:
https://github.com/LauraRuis/groundedSCAN

* For the `spatial_relation_splits`:
https://github.com/google-research/language/tree/master/language/gscan/data
"""

_CITATION = """
@inproceedings{NEURIPS2020_e5a90182,
 author = {Ruis, Laura and Andreas, Jacob and Baroni, Marco and Bouchacourt, Diane and Lake, Brenden M},
 booktitle = {Advances in Neural Information Processing Systems},
 editor = {H. Larochelle and M. Ranzato and R. Hadsell and M. F. Balcan and H. Lin},
 pages = {19861--19872},
 publisher = {Curran Associates, Inc.},
 title = {A Benchmark for Systematic Generalization in Grounded Language Understanding},
 url = {https://proceedings.neurips.cc/paper/2020/file/e5a90182cc81e12ab5e72d66e0b46fe3-Paper.pdf},
 volume = {33},
 year = {2020}
}

@inproceedings{qiu-etal-2021-systematic,
    title = "Systematic Generalization on g{SCAN}: {W}hat is Nearly Solved and What is Next?",
    author = "Qiu, Linlu  and
      Hu, Hexiang  and
      Zhang, Bowen  and
      Shaw, Peter  and
      Sha, Fei",
    booktitle = "Proceedings of the 2021 Conference on Empirical Methods in Natural Language Processing",
    month = nov,
    year = "2021",
    address = "Online and Punta Cana, Dominican Republic",
    publisher = "Association for Computational Linguistics",
    url = "https://aclanthology.org/2021.emnlp-main.166",
    doi = "10.18653/v1/2021.emnlp-main.166",
    pages = "2180--2188",
}
"""

_GSCAN_DATA_PATH = (
    'https://raw.githubusercontent.com/LauraRuis/groundedSCAN/master/data/'
)
_SPATIAL_DATA_PATH = 'https://storage.googleapis.com/gresearch/gscan/'


class GroundedScanConfig(tfds.core.BuilderConfig):
  """BuilderConfig for groundedSCAN."""

  def __init__(self, *, data_path: str, splits_names: List[str], **kwargs):
    """BuilderConfig for groundedSCAN.

    Args:
      data_path: path from which data will be downloaded.
      splits_names: list of splits for the given config.
      **kwargs: keyword arguments forwarded to super.
    """
    super(GroundedScanConfig, self).__init__(**kwargs)
    self.data_path = data_path
    self.splits_names = splits_names


class GroundedScan(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for grounded_scan dataset."""

  BUILDER_CONFIGS = [
      GroundedScanConfig(
          name='compositional_splits',
          description='Examples for compositional generalization.',
          data_path=os.path.join(_GSCAN_DATA_PATH, 'compositional_splits.zip'),
          splits_names=[
              'train',
              'dev',
              'test',
              'visual',
              'situational_1',
              'situational_2',
              'contextual',
              'adverb_1',
              'adverb_2',
              'visual_easier',
          ],
      ),
      GroundedScanConfig(
          name='target_length_split',
          description='Examples for generalizing to larger target lengths.',
          data_path=os.path.join(_GSCAN_DATA_PATH, 'target_length_split.zip'),
          splits_names=['train', 'dev', 'test', 'target_lengths'],
      ),
      GroundedScanConfig(
          name='spatial_relation_splits',
          description='Examples for spatial relation reasoning.',
          data_path=os.path.join(
              _SPATIAL_DATA_PATH, 'spatial_relation_splits.zip'
          ),
          splits_names=[
              'train',
              'dev',
              'test',
              'visual',
              'relation',
              'referent',
              'relative_position_1',
              'relative_position_2',
          ],
      ),
  ]

  VERSION = tfds.core.Version('2.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
      '1.1.0': 'Changed `vector` feature to Text().',
      '2.0.0': 'Adds the new spatial_relation_splits config.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    position_feature = tfds.features.FeaturesDict(
        {'row': np.int32, 'column': np.int32}
    )
    object_feature = tfds.features.FeaturesDict({
        'vector': tfds.features.Text(),
        'position': position_feature,
        'object': tfds.features.FeaturesDict({
            'shape': tfds.features.Text(),
            'color': tfds.features.Text(),
            'size': np.int32,
        }),
    })
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            'situation': tfds.features.FeaturesDict({
                'grid_size': np.int32,
                'agent_position': position_feature,
                'agent_direction': np.int32,
                'target_object': object_feature,
                'distance_to_target': np.int32,
                'direction_to_target': tfds.features.Text(),
                'placed_objects': tfds.features.Sequence(object_feature),
            }),
            'target_commands': tfds.features.Sequence(tfds.features.Text()),
            'command': tfds.features.Sequence(tfds.features.Text()),
            'meaning': tfds.features.Sequence(tfds.features.Text()),
            'verb_in_command': tfds.features.Text(),
            'manner': tfds.features.Text(),
            'referred_target': tfds.features.Text(),
        }),
        supervised_keys=None,
        homepage='https://github.com/LauraRuis/groundedSCAN',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""

    path = dl_manager.download_and_extract(self.builder_config.data_path)
    return {
        split_name: self._generate_examples(
            path / self.builder_config.name, split_name=split_name
        )
        for split_name in self.builder_config.splits_names
    }

  def _generate_examples(self, path, split_name):
    """Yields examples."""

    beam = tfds.core.lazy_imports.apache_beam

    def _get_position_feature(raw_position):
      return {
          'row': int(raw_position['row']),
          'column': int(raw_position['column']),
      }

    def _get_object_feature(raw_object):
      return {
          'vector': raw_object['vector'].strip(),
          'position': _get_position_feature(raw_object['position']),
          'object': {
              'shape': raw_object['object']['shape'],
              'color': raw_object['object']['color'],
              'size': int(raw_object['object']['size']),
          },
      }

    def _parse_sparse_situation_to_feature(situation):
      return {
          'grid_size': int(situation['grid_size']),
          'agent_direction': int(situation['agent_direction']),
          'distance_to_target': int(situation['grid_size']),
          'direction_to_target': situation['direction_to_target'],
          'agent_position': _get_position_feature(situation['agent_position']),
          'target_object': _get_object_feature(situation['target_object']),
          'placed_objects': [
              _get_object_feature(obj)
              for obj in situation['placed_objects'].values()
          ],
      }

    def _preprocess(example):
      return {
          'command': example['command'].split(','),
          'target_commands': example['target_commands'].split(','),
          'meaning': example['meaning'].split(','),
          'manner': example['manner'],
          'verb_in_command': example['verb_in_command'],
          'referred_target': example['referred_target'],
          'situation': _parse_sparse_situation_to_feature(example['situation']),
      }

    def _yield_examples(path):
      dataset_path = os.path.join(path, 'dataset.txt')
      with tf.io.gfile.GFile(dataset_path, 'r') as f:
        dataset = json.load(f)
      for i, example in enumerate(dataset['examples'][split_name]):
        yield f'{split_name}_{i}', _preprocess(example)

    return 'Create pipeline' >> beam.Create(
        [path]
    ) | 'Process samples' >> beam.FlatMap(_yield_examples)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""WMT13: Translate dataset."""

import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.translate import wmt

_URL = "http://www.statmt.org/wmt13/translation-task.html"
_CITATION = """
@InProceedings{bojar-EtAl:2013:WMT,
  author    = {Bojar, Ondrej  and  Buck, Christian  and  Callison-Burch, Chris  and  Federmann, Christian  and  Haddow, Barry  and  Koehn, Philipp  and  Monz, Christof  and  Post, Matt  and  Soricut, Radu  and  Specia, Lucia},
  title     = {Findings of the 2013 {Workshop on Statistical Machine Translation}},
  booktitle = {Proceedings of the Eighth Workshop on Statistical Machine Translation},
  month     = {August},
  year      = {2013},
  address   = {Sofia, Bulgaria},
  publisher = {Association for Computational Linguistics},
  pages     = {1--44},
  url       = {http://www.aclweb.org/anthology/W13-2201}
}

"""

_LANGUAGE_PAIRS = [(lang, "en") for lang in ["cs", "de", "fr", "es", "ru"]]


class Wmt13Translate(wmt.WmtTranslate):
  """WMT 13 translation datasets for all {xx, "en"} language pairs."""

  # Version history:
  # 1.0.0: S3 (new shuffling, sharding and slicing mechanism).
  BUILDER_CONFIGS = [
      wmt.WmtConfig(  # pylint:disable=g-complex-comprehension
          description="WMT 2013 %s-%s translation task dataset." % (l1, l2),
          url=_URL,
          citation=_CITATION,
          language_pair=(l1, l2),
          version=tfds.core.Version("1.0.0"),
      )
      for l1, l2 in _LANGUAGE_PAIRS
  ]

  @property
  def _subsets(self):
    return {
        tfds.Split.TRAIN: [
            "europarl_v7",
            "commoncrawl",
            "multiun",
            "newscommentary_v8",
            "gigafren",
            "wikiheadlines_ru",
            "yandexcorpus",
            "czeng_10",
        ],
        tfds.Split.VALIDATION: [
            "newstest2012",
            "newstest2011",
            "newstest2010",
            "newstest2009",
            "newstest2008",
            "newssyscomb2009",
        ],
        tfds.Split.TEST: ["newstest2013"],
    }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logic to read sharded files (tfrecord, buckets, ...).

This logic is shared between:
 - tfrecord_reader, to read sharded tfrecord files, based on user instructions.
 - tfrecord_writer, to read sharded bucket files (temp files), based on final
 sharding needs.
"""
from __future__ import annotations

import dataclasses
import math
import os
from typing import Any, List, Optional, Sequence

DEFAULT_MIN_SHARD_SIZE: int = 64 << 20  # 64 MiB
DEFAULT_MAX_SHARD_SIZE: int = 1024 << 20  # 1 GiB


@dataclasses.dataclass(frozen=True)
class ShardConfig:
  """Configuration of how shards should be created.

  Attributes:
    num_shards: number of shards that should be used. If `None`, then the number
      of shards is computed based on the total size of the dataset and the min
      and max shard size.
    min_shard_size: minimum shard size in bytes.
    max_shard_size: maximum shard size in bytes.
    overhead: the amount of overhead when writing a file. By default, it is the
      TFRecord overhead. See
      https://github.com/tensorflow/tensorflow/blob/27325fabed898880fa1b33a04d4b125a6ef4bbc8/tensorflow/core/lib/io/record_writer.h#L104
  """

  num_shards: Optional[int] = None
  min_shard_size: int = DEFAULT_MIN_SHARD_SIZE
  max_shard_size: int = DEFAULT_MAX_SHARD_SIZE
  overhead: int = 16

  def get_number_shards(
      self,
      total_size: int,
      num_examples: int,
      uses_precise_sharding: bool = True,
  ) -> int:
    """Returns number of shards for num_examples of total_size in bytes.

    Each shard should be at least 128MB.
    A pod has 16*16=256 TPU devices containing 1024 TPU chips (2048 cores).
    So if the dataset is large enough, we want the number of shards to be a
    multiple of 1024, but with shards as big as possible.
    If the dataset is too small, we want the number of shards to be a power
    of two so it distributes better on smaller TPU configs (8, 16, 32, ...
    cores).

    Args:
      total_size: the size of the data (serialized, not couting any overhead).
      num_examples: the number of records in the data.
      uses_precise_sharding: whether a mechanism is used to exactly control how
        many examples go in each shard.

    Returns:
      number of shards to use.
    """
    if self.num_shards:
      return self.num_shards

    total_size += num_examples * self.overhead
    max_shards_number = total_size // self.min_shard_size
    if uses_precise_sharding:
      max_shard_size = self.max_shard_size
    else:
      # When the pipeline does not control exactly how many rows go into each
      # shard (called 'precise sharding' here), we use a smaller max shard size
      # so that the pipeline doesn't fail if a shard gets some more examples.
      max_shard_size = 0.9 * self.max_shard_size
    min_shards_number = total_size // max_shard_size
    if min_shards_number <= 1024 <= max_shards_number and num_examples >= 1024:
      return 1024
    elif min_shards_number > 1024:
      i = 2
      while True:
        n = 1024 * i
        if n >= min_shards_number and num_examples >= n:
          return n
        i += 1
    else:
      for n in [512, 256, 128, 64, 32, 16, 8, 4, 2]:
        if min_shards_number <= n <= max_shards_number and num_examples >= n:
          return n
    return 1


@dataclasses.dataclass(eq=True, frozen=True)
class FileInstruction:
  """Instruction to read a single shard/file.

  Attributes:
    filename: The filename including the path.
    skip: Indicates which example read in the shard (`ds.skip().take()`). `0` if
      no skipping.
    take: Indicates how many examples to read (`-1` to read all). If `-1`, then
      `take` is overridden with the number of examples to read is calculated.
    examples_in_shard: The total number of examples in the shard. To get the
      number of examples to read, use `take`.
    takes_all: whether all examples after skipping are read.
  """

  filename: str
  skip: int
  take: int
  examples_in_shard: int

  def __post_init__(self):
    if self.take == -1:
      object.__setattr__(self, 'take', self.examples_in_shard - self.skip)
    if self.examples_in_shard < 0:
      raise ValueError(
          f'examples_in_shard should be >= 0! Was {self.examples_in_shard}.'
      )
    if self.skip < 0 or self.skip > self.examples_in_shard:
      raise ValueError(
          f'skip should be between 0 and {self.examples_in_shard}! '
          f'Was {self.skip}.'
      )
    if self.skip + self.take > self.examples_in_shard:
      raise ValueError(
          f'skip ({self.skip}) + take ({self.take}) should be '
          f'<= examples_in_shard ({self.examples_in_shard})!'
      )

  @property
  def takes_all(self) -> bool:
    return self.skip + self.take == self.examples_in_shard

  def dirname(self) -> str:
    return os.path.dirname(self.filename)

  def basename(self) -> str:
    return os.path.basename(self.filename)

  def replace(self, **kwargs: Any) -> FileInstruction:
    return dataclasses.replace(self, **kwargs)


def split_file_instruction(
    file_instruction: FileInstruction,
    num_splits: int,
) -> List[FileInstruction]:
  """Instructions for reading the given file instruction in several splits.

  Note that this function may return fewer splits than `num_splits` in case the
  number of examples cannot be split into that many. For example, if
  `file_instruction` has `take=1` and `num_splits=2`, then only a single
  file instruction is returned.

  Arguments:
    file_instruction: the file instruction that should be split into multiple
      file instructions. It is currently not supported to have `skip>0` or not
      take all examples in the file.
    num_splits: the number of splits the file instruction should be split into.

  Returns:
    list of file instructions into which it is split.
  """
  if not file_instruction.takes_all or file_instruction.skip > 0:
    raise ValueError('Only file instructions that read all rows are supported!')
  examples_per_split = math.ceil(file_instruction.take / num_splits)
  splits = []
  index = file_instruction.skip
  while index < file_instruction.examples_in_shard:
    # If there are fewer examples left in the file than `examples_per_split`,
    # then only take what's left.
    take = min(examples_per_split, file_instruction.examples_in_shard - index)
    splits.append(file_instruction.replace(skip=index, take=take))
    index += take
  return splits


def get_file_instructions(
    from_: int,
    to: int,
    filenames: Sequence[str],
    shard_lengths: Sequence[int],
) -> List[FileInstruction]:
  """Returns a list of files (+skip/take) to read [from_:to] items from shards.

  Args:
    from_: int, Index (included) of element from which to read.
    to: int, Index (excluded) of element to which to read.
    filenames: list of strings or ints, the filenames of the shards. Not really
      used, but to place in result.
    shard_lengths: the number of elements in every shard.

  Returns:
    list of dict(filename, skip, take).
  """
  index_start = 0  # Beginning (included) of moving window.
  index_end = 0  # End (excluded) of moving window.
  file_instructions = []
  for filename, length in zip(filenames, shard_lengths):
    if not length:
      continue  # Empty shard - can happen with temporary buckets.
    index_end += length
    if from_ < index_end and to > index_start:  # There is something to take.
      skip = from_ - index_start if from_ > index_start else 0
      take = to - index_start - skip if to < index_end else -1
      if take == 0:
        continue
      file_instructions.append(
          FileInstruction(
              filename=filename, skip=skip, take=take, examples_in_shard=length
          )
      )
    index_start += length
  return file_instructions



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TFDS typing annotations."""

from etils import epath
from tensorflow_datasets.core import decode
from tensorflow_datasets.core import splits
from tensorflow_datasets.core.features import feature
from tensorflow_datasets.core.utils import type_utils

# pylint: disable=unused-import
from tensorflow_datasets.core.utils.type_utils import *  # pylint: disable=wildcard-import
# pylint: enable=unused-import

# Accept both `str` and `pathlib.Path`-like
PathLike = epath.PathLike

DecoderArg = decode.partial_decode.DecoderArg
FeatureSpecs = decode.partial_decode.FeatureSpecs
FeatureConnectorArg = feature.FeatureConnectorArg
SplitArg = splits.SplitArg

__all__ = type_utils.__all__ + [
    'DecoderArg',
    'FeatureConnectorArg',
    'FeatureSpecs',
    'PathLike',
    'SplitArg',
]



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""mt_opt dataset."""

from __future__ import annotations

import os
from typing import Any, Dict, Generator, Tuple

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds


_DESCRIPTION = """
Datasets for the [MT-Opt paper](https://arxiv.org/abs/2104.08212).
"""

_CITATION = """
@misc{kalashnikov2021mtopt,
      title={MT-Opt: Continuous Multi-Task Robotic Reinforcement Learning at Scale},
      author={Dmitry Kalashnikov and Jacob Varley and Yevgen Chebotar and Benjamin Swanson and Rico Jonschkowski and Chelsea Finn and Sergey Levine and Karol Hausman},
      year={2021},
      eprint={2104.08212},
      archivePrefix={arXiv},
      primaryClass={cs.RO}
}
"""

_BUILDER_CONFIGS = [
    tfds.core.BuilderConfig(
        name='rlds',
        description=(
            'This dataset contains task episodes collected across afleet of'
            ' real robots. It follows the [RLDS'
            ' format](https://github.com/google-research/rlds)to represent'
            ' steps and episodes.'
        ),
    ),
    tfds.core.BuilderConfig(
        name='sd',
        description=(
            'The success detectors dataset that contains human curated'
            ' definitions of tasks completion.'
        ),
    ),
]


def _steps_features():
  return tfds.features.FeaturesDict({
      'action': tfds.features.FeaturesDict({
          'close_gripper': np.bool_,
          'open_gripper': np.bool_,
          'target_pose': tfds.features.Tensor(
              shape=(7,), dtype=np.float32, encoding=tfds.features.Encoding.ZLIB
          ),
          'terminate': np.bool_,
      }),
      'is_first': np.bool_,
      'is_last': np.bool_,
      'is_terminal': np.bool_,
      'observation': tfds.features.FeaturesDict({
          'gripper_closed': np.bool_,
          'height_to_bottom': np.float32,
          'image': tfds.features.Image(shape=(512, 640, 3), dtype=np.uint8),
          'state_dense': tfds.features.Tensor(
              shape=(7,), dtype=np.float32, encoding=tfds.features.Encoding.ZLIB
          ),
      }),
  })


def _name_to_features(config_name: str):
  if config_name == 'rlds':
    return tfds.features.FeaturesDict({
        'episode_id': np.str_,
        'skill': np.uint8,
        'steps': tfds.features.Dataset(_steps_features()),
        'task_code': np.str_,
    })
  return tfds.features.FeaturesDict({
      'image_0': tfds.features.Image(shape=(512, 640, 3), dtype=np.uint8),
      'image_1': tfds.features.Image(shape=(480, 640, 3), dtype=np.uint8),
      'image_2': tfds.features.Image(shape=(480, 640, 3), dtype=np.uint8),
      'success': np.bool_,
      'task_code': np.str_,
  })


# To encode, we use sequence instead of nested dataset. Otherwise, Beam has
# issues calculating the size of the yielded examples (b/219881125)
def _name_to_features_encode(config_name: str):
  if config_name == 'rlds':
    return tfds.features.FeaturesDict({
        'episode_id': np.str_,
        'skill': np.uint8,
        'steps': tfds.features.Sequence(_steps_features()),
        'task_code': np.str_,
    })
  return tfds.features.FeaturesDict({
      'image_0': tfds.features.Image(shape=(512, 640, 3), dtype=np.uint8),
      'image_1': tfds.features.Image(shape=(480, 640, 3), dtype=np.uint8),
      'image_2': tfds.features.Image(shape=(480, 640, 3), dtype=np.uint8),
      'success': np.bool_,
      'task_code': np.str_,
  })


_NAME_TO_SPLITS = {
    'sd': {
        'train': 1024,
        'test': 256,
    },
    'rlds': {
        'train': 2048,
    },
}


def _filename(prefix: str, num_shards: int, shard_id: int):
  return os.fspath(
      tfds.core.Path(f'{prefix}-{shard_id:05d}-of-{num_shards:05d}')
  )


def _get_files(prefix: str, ds_name: str, split: str, num_shards: int):
  prefix = f'{prefix}/mt_opt_{ds_name}/1.0.0/mt_opt_{ds_name}-{split}.tfrecord'
  return [_filename(prefix, num_shards, i) for i in range(num_shards)]


class MtOpt(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for mt_opt datasets."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }
  BUILDER_CONFIGS = _BUILDER_CONFIGS
  _INPUT_FILE_PREFIX = 'gs://gresearch/robotics/'


  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=_name_to_features(self.builder_config.name),
        supervised_keys=None,
        homepage='https://karolhausman.github.io/mt-opt/',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    ds_name = self.builder_config.name
    splits = {}
    for split, shards in _NAME_TO_SPLITS[ds_name].items():
      paths = {
          'file_paths': _get_files(
              self._INPUT_FILE_PREFIX, ds_name, split, shards
          )
      }
      splits[split] = self._generate_examples(paths)
    return splits

  def _generate_examples_one_file(
      self, path
  ) -> Generator[Tuple[str, Dict[str, Any]], None, None]:
    """Yields examples from one file."""
    # Dataset of tf.Examples containing full episodes.
    example_ds = tf.data.TFRecordDataset(filenames=str(path))

    example_features = _name_to_features_encode(self.builder_config.name)
    example_specs = example_features.get_serialized_info()
    parser = tfds.core.example_parser.ExampleParser(example_specs)

    parsed_examples = example_ds.map(parser.parse_example)
    decoded_examples = parsed_examples.map(example_features.decode_example)

    for index, example in enumerate(tfds.as_numpy(decoded_examples)):
      if self.builder_config.name == 'rlds':
        id_key = 'episode_id'
      else:
        id_key = 'task_code'
      example_id = str(index) + str(example[id_key]) + str(hash(path))
      yield example_id, example

  def _generate_examples(self, paths):
    """Yields examples."""
    beam = tfds.core.lazy_imports.apache_beam
    file_paths = paths['file_paths']

    return beam.Create(file_paths) | beam.FlatMap(
        self._generate_examples_one_file
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Librispeech dataset."""

import csv
import os

from etils import epath
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

# TODO(b/228460133): Update with pages and doi when available.
_CITATION = """\
@inproceedings{breiner2022userlibri,
  title={UserLibri: A Dataset for ASR Personalization Using Only Text},
  author={Breiner, Theresa and Ramaswamy, Swaroop and Variani, Ehsan and Garg, Shefali and Mathews, Rajiv and Sim, Khe Chai and Gupta, Kilol and Chen, Mingqing and McConnaughey, Lara},
  booktitle={Proc. Interspeech 2022},
  year={2022}
}
"""

_DESCRIPTION = """\
UserLibri is a dataset containing paired audio-transcripts and additional text
only data for each of 107 users. It is a reformatting of the LibriSpeech dataset
found at http://www.openslr.org/12, reorganizing the data into users with an
average of 52 LibriSpeech utterances and about 6,700 text example sentences per
user. The UserLibriAudio class provides access to the audio-transcript pairs.
See UserLibriText for the additional text data.
"""

_URL = "https://www.kaggle.com/datasets/google/userlibri"
_DL_URL = tfds.download.Resource(
    url="https://www.kaggle.com/datasets/google/userlibri/download",
    extract_method=tfds.download.ExtractMethod.ZIP,
)

_KAGGLE_DATASET_ID = "google/userlibri"


def read_metadata_file(path):
  """Reads the tab-separated metadata from the path."""
  metadata = {}
  with epath.Path(path).open() as f:
    reader = csv.DictReader(f, delimiter="\t")
    for row in reader:
      # Collect metadata for each split_userID, for example
      # test-other_speaker-12-book-34.
      metadata[f'{row["Split"]}_{row["User ID"]}'] = {
          "Num Audio Examples": row["Num Audio Examples"],
          "Average Words Per Example": row["Average Words Per Example"],
      }
  return metadata


class UserLibriAudio(tfds.core.BeamBasedBuilder):
  """UserLibri dataset of paired audio transcripts."""

  VERSION = tfds.core.Version("1.0.0")

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            "id": tfds.features.Text(doc="The id of this unique utterance"),
            "user_id": tfds.features.Text(
                doc=(
                    "The user that this utterance belongs to (unique speaker"
                    " and book)"
                )
            ),
            "speaker_id": tfds.features.Text(
                doc="The speaker who read this utterance"
            ),
            "book_id": tfds.features.Text(
                doc="The book that this utterance is read from"
            ),
            "audio": tfds.features.Audio(
                sample_rate=16000,
                doc=(
                    "The audio clip containing a snippet from a book read aloud"
                ),
            ),
            "transcript": tfds.features.Text(
                doc="The text that the speaker read to produce the audio"
            ),
        }),
        supervised_keys=("audio", "transcript"),
        homepage=_URL,
        citation=_CITATION,
        metadata=tfds.core.MetadataDict(
            sample_rate=16000,
        ),
    )

  def _split_generators(self, dl_manager):
    """Generates user splits like 'test-other_speaker-1234-book-5678'."""
    extracted_dir = dl_manager.download_kaggle_data(
        competition_or_dataset=_KAGGLE_DATASET_ID
    )
    audio_dir = extracted_dir / "UserLibri/audio_data"
    # Load metadata file which is the same for both test-clean and test-other.
    metadata_file = audio_dir / "metadata.tsv"
    self.info.metadata["users"] = read_metadata_file(metadata_file)
    splits = []
    for split in ["test-clean", "test-other"]:
      split_path = audio_dir / split
      for user_id in tf.io.gfile.listdir(split_path):
        split_and_user = f"{split}_{user_id}"
        splits.append(
            tfds.core.SplitGenerator(
                name=split_and_user,
                gen_kwargs={
                    "directory": os.path.join(audio_dir, split, user_id)
                },
            )
        )
    return splits

  def _build_pcollection(self, pipeline, directory):
    """Generates examples as dicts."""
    beam = tfds.core.lazy_imports.apache_beam
    return (
        pipeline
        | beam.Create([directory])
        | beam.FlatMap(_generate_userlibri_audio_examples)
        | beam.Reshuffle()
    )


def _generate_userlibri_audio_examples(user_directory):
  """Generate audio examples from a UserLibri directory."""
  transcripts_glob = os.path.join(user_directory, "*.txt")
  for transcript_file in tf.io.gfile.glob(transcripts_glob):
    with tf.io.gfile.GFile(os.path.join(user_directory, transcript_file)) as f:
      for line in f:
        line = line.strip()
        key, transcript = line.split(" ", 1)
        audio_file = f"{key}.flac"
        user_id = user_directory.split("/")[-1]
        _, speaker_id, _, book_id = user_id.split("-")
        example = {
            "id": key,
            "user_id": user_id,
            "speaker_id": speaker_id,
            "book_id": book_id,
            "audio": os.path.join(user_directory, audio_file),
            "transcript": transcript,
        }
        yield key, example



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Transform API."""

from tensorflow_datasets.core.transform.transform_lib import apply_do_fn
from tensorflow_datasets.core.transform.transform_lib import apply_filter
from tensorflow_datasets.core.transform.transform_lib import apply_fn
from tensorflow_datasets.core.transform.transform_lib import apply_transformations
from tensorflow_datasets.core.transform.transform_lib import Example
from tensorflow_datasets.core.transform.transform_lib import ExampleTransformFn
from tensorflow_datasets.core.transform.transform_lib import Key
from tensorflow_datasets.core.transform.transform_lib import KeyExample
from tensorflow_datasets.core.transform.transform_lib import remove_feature
from tensorflow_datasets.core.transform.transform_lib import rename_feature
from tensorflow_datasets.core.transform.transform_lib import rename_features

__all__ = [
    "apply_do_fn",
    "apply_filter",
    "apply_fn",
    "apply_transformations",
    "Example",
    "ExampleTransformFn",
    "Key",
    "KeyExample",
    "remove_feature",
    "rename_feature",
    "rename_features",
]



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Error message helper for TFDS."""

import contextlib
import dataclasses
import threading
from typing import Callable, Iterator, List, Type, Union

from tensorflow_datasets.core import utils

Message = Union[str, Callable[[], str]]


@dataclasses.dataclass
class ErrorContext:
  """Stack container for error context.

  This stack keeps track of the error messages which are raised when loading a
  dataset. These error messages are used to provide better context for the users
  in case of DatasetNotFound errors are raised.
  """

  messages: List[Message] = dataclasses.field(default_factory=list)


# Current error context. Accessed by `reraise_with_context` and `add_context`.
context_holder = threading.local()


@contextlib.contextmanager
def reraise_with_context(error_cls: Type[Exception]) -> Iterator[None]:
  """Contextmanager which reraises an exception with an additional message.

  Args:
    error_cls: The exception to be reraised.

  Yields:
    None.
  """
  # If current_context_msg exists, we are already within the scope of the
  # session contextmanager.
  if hasattr(context_holder, 'current_context_msg'):
    yield
    return

  context_holder.current_context_msg = ErrorContext()
  try:
    yield
  except error_cls as e:
    context_msg = '\n'.join(context_holder.current_context_msg.messages)
    utils.reraise(e, suffix=context_msg)
  finally:
    del context_holder.current_context_msg


def add_context(msg: str) -> None:
  """Appends the error message to the error context stack.

  Args:
    msg: The error message to add to the error context stack.

  Returns:
    None.

  Raises:
    AttributeError if local thread has no current_context_msg attribute.
  """
  if not hasattr(context_holder, 'current_context_msg'):
    raise AttributeError(
        'add_context called outside of reraise_with_context contextmanager.'
    )
  context_holder.current_context_msg.messages.append(msg)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""`tfds build` command."""

import argparse
import functools
import importlib
import itertools
import json
import multiprocessing
import os
import typing
from typing import Dict, Iterator, Optional, Tuple, Type, Union

from absl import logging
from etils import epath
import tensorflow_datasets as tfds

# pylint: disable=logging-fstring-interpolation


def register_subparser(parsers: argparse._SubParsersAction) -> None:  # pylint: disable=protected-access
  """Add subparser for `build` command."""
  build_parser = parsers.add_parser(
      'build', help='Commands for downloading and preparing datasets.'
  )
  build_parser.add_argument(
      'datasets',  # Positional arguments
      type=str,
      nargs='*',
      help=(
          'Name(s) of the dataset(s) to build. Default to current dir. '
          'See https://www.tensorflow.org/datasets/cli for accepted values.'
      ),
  )
  build_parser.add_argument(  # Also accept keyword arguments
      '--datasets',
      type=str,
      nargs='+',
      dest='datasets_keyword',
      help='Datasets can also be provided as keyword argument.',
  )

  # **** Debug options ****
  debug_group = build_parser.add_argument_group(
      'Debug & tests',
      description=(
          '--pdb Enter post-mortem debugging mode if an exception is raised.'
      ),
  )
  debug_group.add_argument(
      '--overwrite',
      action='store_true',
      help='Delete pre-existing dataset if it exists.',
  )
  debug_group.add_argument(
      '--fail_if_exists',
      action='store_true',
      default=False,
      help='Fails the program if there is a pre-existing dataset.',
  )
  debug_group.add_argument(
      '--max_examples_per_split',
      type=int,
      nargs='?',
      const=1,
      help=(
          'When set, only generate the first X examples (default to 1), rather'
          ' than the full dataset.If set to 0, only execute the'
          ' `_split_generators` (which download the original data), but skip'
          ' `_generator_examples`'
      ),
  )

  # **** Path options ****
  path_group = build_parser.add_argument_group('Paths')
  path_group.add_argument(
      '--data_dir',
      type=tfds.core.Path,
      # Should match tfds.core.constant.DATA_DIR !!
      default=tfds.core.Path(
          os.environ.get(
              'TFDS_DATA_DIR',
              os.path.join(os.path.expanduser('~'), 'tensorflow_datasets'),
          )
      ),
      help=(
          'Where to place datasets. Default to '
          '`~/tensorflow_datasets/` or `TFDS_DATA_DIR` environement variable.'
      ),
  )
  path_group.add_argument(
      '--download_dir',
      type=tfds.core.Path,
      help='Where to place downloads. Default to `<data_dir>/downloads/`.',
  )
  path_group.add_argument(
      '--extract_dir',
      type=tfds.core.Path,
      help='Where to extract files. Default to `<download_dir>/extracted/`.',
  )
  path_group.add_argument(
      '--manual_dir',
      type=tfds.core.Path,
      help=(
          'Where to manually download data (required for some datasets). '
          'Default to `<download_dir>/manual/`.'
      ),
  )
  path_group.add_argument(
      '--add_name_to_manual_dir',
      action='store_true',
      help=(
          'If true, append the dataset name to the `manual_dir` (e.g. '
          '`<download_dir>/manual/<dataset_name>/`. Useful to avoid collisions '
          'if many datasets are generated.'
      ),
  )

  # **** Generation options ****
  generation_group = build_parser.add_argument_group('Generation')
  generation_group.add_argument(
      '--download_only',
      action='store_true',
      help=(
          'If True, download all files but do not prepare the dataset. Uses the'
          ' checksum.tsv to find out what to download. Therefore, this does not'
          ' work in combination with --register_checksums.'
      ),
  )
  generation_group.add_argument(
      '--config',
      '-c',
      type=str,
      help=(
          'Config name to build. Build all configs if not set. Can also be a'
          ' json of the kwargs forwarded to the config `__init__` (for custom'
          ' configs).'
      ),
  )
  # We are forced to have 2 flags to avoid ambiguity when config name is
  # a number (e.g. `voc/2017`)
  generation_group.add_argument(
      '--config_idx',
      type=int,
      help=(
          'Config id to build (`builder_cls.BUILDER_CONFIGS[config_idx]`). '
          'Mutually exclusive with `--config`.'
      ),
  )
  generation_group.add_argument(
      '--update_metadata_only',
      action='store_true',
      default=False,
      help=(
          'If True, existing dataset_info.json is updated with metadata defined'
          ' in Builder class(es). Datasets must already have been prepared.'
      ),
  )
  generation_group.add_argument(
      '--download_config',
      type=str,
      help=(
          'A json of the kwargs forwarded to the config `__init__` (for custom'
          ' DownloadConfigs).'
      ),
  )
  generation_group.add_argument(
      '--imports',
      '-i',
      type=str,
      help='Comma separated list of module to import to register datasets.',
  )
  generation_group.add_argument(
      '--register_checksums',
      action='store_true',
      help='If True, store size and checksum of downloaded files.',
  )
  generation_group.add_argument(
      '--force_checksums_validation',
      action='store_true',
      help='If True, raise an error if the checksums are not found.',
  )
  generation_group.add_argument(
      '--beam_pipeline_options',
      type=str,
      # nargs='+',
      help=(
          'A (comma-separated) list of flags to pass to `PipelineOptions` when'
          ' preparing with Apache Beam. (see:'
          ' https://www.tensorflow.org/datasets/beam_datasets). Example:'
          ' `--beam_pipeline_options=job_name=my-job,project=my-project`'
      ),
  )
  format_values = [f.value for f in tfds.core.FileFormat]
  generation_group.add_argument(
      '--file_format',
      type=str,
      help=(
          'File format to which generate the tf-examples. '
          f'Available values: {format_values} (see `tfds.core.FileFormat`).'
      ),
  )
  generation_group.add_argument(
      '--max_shard_size_mb', type=int, help='The max shard size in megabytes.'
  )
  generation_group.add_argument(
      '--num-processes',
      type=int,
      default=1,
      help='Number of parallel build processes.',
  )

  publish_group = build_parser.add_argument_group(
      'Publishing',
      description='Options for publishing successfully created datasets.',
  )
  publish_group.add_argument(
      '--publish_dir',
      type=tfds.core.Path,
      default=None,
      required=False,
      help=(
          'Where to optionally publish the dataset after it has been '
          'generated successfully. Should be the root data dir under which'
          'datasets are stored. '
          'If unspecified, dataset will not be published'
      ),
  )
  publish_group.add_argument(
      '--skip_if_published',
      action='store_true',
      default=False,
      help=(
          'If the dataset with the same version and config is already '
          'published, then it will not be regenerated.'
      ),
  )

  # **** Automation options ****
  automation_group = build_parser.add_argument_group(
      'Automation', description='Used by automated scripts.'
  )
  automation_group.add_argument(
      '--exclude_datasets',
      type=str,
      help=(
          'If set, generate all datasets except the one defined here. '
          'Comma separated list of datasets to exclude. '
      ),
  )
  automation_group.add_argument(
      '--experimental_latest_version',
      action='store_true',
      help=(
          'Build the latest Version(experiments=...) available rather than '
          'default version.'
      ),
  )

  build_parser.set_defaults(subparser_fn=_build_datasets)


def _build_datasets(args: argparse.Namespace) -> None:
  """Build the given datasets."""
  # Eventually register additional datasets imports
  if args.imports:
    list(importlib.import_module(m) for m in args.imports.split(','))

  # Select datasets to generate
  datasets = (args.datasets or []) + (args.datasets_keyword or [])
  if args.exclude_datasets:  # Generate all datasets if `--exclude_datasets` set
    if datasets:
      raise ValueError("--exclude_datasets can't be used with `datasets`")
    datasets = set(tfds.list_builders(with_community_datasets=False)) - set(
        args.exclude_datasets.split(',')
    )
    datasets = sorted(datasets)  # `set` is not deterministic
  else:
    datasets = datasets or ['']  # Empty string for default

  # Parallelize datasets generation.
  builders = itertools.chain(
      *(_make_builders(args, dataset) for dataset in datasets)
  )
  process_builder_fn = functools.partial(
      _download if args.download_only else _download_and_prepare, args
  )

  if args.num_processes == 1:
    for builder in builders:
      process_builder_fn(builder)
  else:
    with multiprocessing.Pool(args.num_processes) as pool:
      pool.map(process_builder_fn, builders)


def _make_builders(
    args: argparse.Namespace,
    ds_to_build: str,
) -> Iterator[tfds.core.DatasetBuilder]:
  """Yields builders to generate."""
  builder_cls, builder_kwargs = _get_builder_cls(
      ds_to_build,
      has_imports=bool(args.imports),
  )

  # Eventually overwrite version
  if args.experimental_latest_version:
    if 'version' in builder_kwargs:
      raise ValueError(
          "Can't have both `--experimental_latest` and version set (`:1.0.0`)"
      )
    builder_kwargs['version'] = 'experimental_latest'

  # Eventually overwrite config
  builder_kwargs['config'] = _get_config_name(
      builder_cls=builder_cls,
      config_kwarg=builder_kwargs.get('config'),
      config_name=args.config,
      config_idx=args.config_idx,
  )

  if args.file_format:
    builder_kwargs['file_format'] = args.file_format

  make_builder = functools.partial(
      _make_builder,
      builder_cls,
      overwrite=args.overwrite,
      fail_if_exists=args.fail_if_exists,
      data_dir=args.data_dir,
      **builder_kwargs,
  )

  # Generate all configs if no config requested.
  if builder_cls.BUILDER_CONFIGS and builder_kwargs['config'] is None:
    for config in builder_cls.BUILDER_CONFIGS:
      yield make_builder(config=config.name)
  # Generate only the dataset
  else:
    yield make_builder()


def _get_builder_cls(
    ds_to_build: str,
    *,
    has_imports: bool,
) -> Tuple[Type[tfds.core.DatasetBuilder], Dict[str, str]]:
  """Infer the builder class to build.

  Args:
    ds_to_build: Dataset argument.
    has_imports: Whether `--imports` was passed

  Returns:
    builder_cls: The dataset class to download and prepare
    kwargs:
  """
  # 1st case: Requested dataset is a path to `.py` script
  # When `--imports=` is set, it means the user expects the dataset to be
  # registered through the `imports` and not locally.
  if not has_imports:
    path = _search_script_path(ds_to_build)
    if path is not None:
      logging.info(f'Loading dataset {ds_to_build} from path: {path}')
      # Dynamically load user dataset script
      # When possible, load from the parent's parent, so module is named
      # "foo.foo_dataset_builder".
      try:
        with tfds.core.utils.add_sys_path(path.parent.parent):
          builder_cls = tfds.core.community.builder_cls_from_module(
              f'{path.parent.stem}.{path.stem}'
          )
        return builder_cls, {}
      except ImportError:
        pass
      # There might be cases where user imports from a legacy builder module,
      # in which case the above wouldn't necessarily work, so we fall back to
      # importing the module directly.
      with tfds.core.utils.add_sys_path(path.parent):
        builder_cls = tfds.core.community.builder_cls_from_module(path.stem)
      return builder_cls, {}

  # 2nd case: Dataset is registered through imports.

  # Extract `name/config:version`
  name, builder_kwargs = tfds.core.naming.parse_builder_name_kwargs(ds_to_build)
  builder_cls = tfds.builder_cls(str(name))
  logging.info(
      f'Loading dataset {ds_to_build} from imports: {builder_cls.__module__}'
  )
  builder_kwargs = typing.cast(Dict[str, str], builder_kwargs)
  return builder_cls, builder_kwargs


def _search_script_path(ds_to_build: str) -> Optional[tfds.core.Path]:
  """Check whether the requested dataset match a file on disk."""
  # If the dataset file exists, use it. Valid values are:
  # * Empty string (use default directory)
  # * Dataset folder (e.g. `my_dataset/my_dataset_dataset_builder.py`)
  # * Legacy dataset folder (e.g. `my_dataset/my_dataset.py`)
  # * Dataset file (e.g. `my_dataset.py`)
  # * Dataset module (e.g. `my_dataset`)

  path = tfds.core.Path(ds_to_build).expanduser().resolve()
  if path.exists():
    if path.is_dir():
      # dataset-as-folder, use `my_dataset/my_dataset_dataset_builder.py` ...
      module_path = path / f'{path.name}_dataset_builder.py'
      if module_path.exists():
        return _validate_script_path(module_path)
      # ... or `my_dataset/my_dataset.py` (legacy)
      return _validate_script_path(path / f'{path.name}.py')
    else:  # `my_dataset.py` script
      return _validate_script_path(path)

  # Try with `.py`
  path = path.with_suffix('.py')
  if path.exists():
    return path
  # Path not found. Use heuristic to detect if user intended to pass a path:
  elif (
      not ds_to_build
      or ds_to_build.endswith((os.sep, '.py'))  # ds.py
      or tfds.core.Path(ds_to_build).is_absolute()  # /path/to
      or ds_to_build.count(os.sep) > 1  # path/dataset/config
  ):
    raise FileNotFoundError(
        f'Could not find generation script for `{ds_to_build}`: '
        f'{path} not found.'
    )
  else:
    return None


def _validate_script_path(
    path: tfds.core.Path,
) -> Optional[tfds.core.Path]:
  """Validates and returns the `dataset_builder.py` generation script path."""
  if path.suffix != '.py':
    raise ValueError(f'Expected `.py` file. Invalid dataset path: {path}.')
  if not path.exists():
    raise FileNotFoundError(
        f'Could not find dataset generation script: {path}. '
    )
  return path


def _make_builder(
    builder_cls: Type[tfds.core.DatasetBuilder],
    overwrite: bool,
    fail_if_exists: bool,
    **builder_kwargs,
) -> tfds.core.DatasetBuilder:
  """Builder factory, eventually deleting pre-existing dataset."""
  builder = builder_cls(**builder_kwargs)  # pytype: disable=not-instantiable
  data_exists = builder.data_path.exists()
  if fail_if_exists and data_exists:
    raise RuntimeError(
        'The `fail_if_exists` flag was True and '
        f'the data already exists in {builder.data_path}'
    )
  if overwrite and data_exists:
    builder.data_path.rmtree()  # Delete pre-existing data
    # Re-create the builder with clean state
    builder = builder_cls(**builder_kwargs)  # pytype: disable=not-instantiable
  return builder


def _download(
    args: argparse.Namespace,
    builder: tfds.core.DatasetBuilder,
) -> None:
  """Downloads all files of the given builder."""
  dl_config = _make_download_config(args, dataset_name=builder.name)

  if dl_config.register_checksums:
    raise ValueError(
        'It is not allowed to enable --register_checksums '
        'when using --download_only!'
    )

  if builder.url_infos is None:
    raise ValueError(
        f'URL infos of builder {builder.name} are not defined, '
        'so cannot download these URLs.'
    )

  max_simultaneous_downloads = None
  if builder.MAX_SIMULTANEOUS_DOWNLOADS is not None:
    max_simultaneous_downloads = builder.MAX_SIMULTANEOUS_DOWNLOADS

  download_dir = args.download_dir or os.path.join(
      builder._data_dir_root, 'downloads'  # pylint: disable=protected-access
  )
  dl_manager = tfds.download.DownloadManager(
      download_dir=download_dir,
      url_infos=builder.url_infos,
      force_download=False,
      force_extraction=False,
      force_checksums_validation=dl_config.force_checksums_validation,
      register_checksums=False,
      verify_ssl=dl_config.verify_ssl,
      dataset_name=builder.name,
      max_simultaneous_downloads=max_simultaneous_downloads,
  )

  urls_to_download = {
      f'file{i}': url for i, url in enumerate(builder.url_infos.keys())
  }
  dl_manager.download(urls_to_download)


def _download_and_prepare(
    args: argparse.Namespace,
    builder: tfds.core.DatasetBuilder,
) -> None:
  """Generate a single builder."""
  logging.info(f'download_and_prepare for dataset {builder.info.full_name}...')

  publish_data_dir = _publish_data_dir(args.publish_dir, builder)
  if args.skip_if_published and publish_data_dir is not None:
    if publish_data_dir.exists():
      logging.info(
          f'Dataset already exists in {publish_data_dir}. Skipping generation.'
      )
      return

  dl_config = _make_download_config(args, dataset_name=builder.name)
  builder.download_and_prepare(
      download_dir=args.download_dir,
      download_config=dl_config,
  )

  # Dataset generated successfully
  logging.info('Dataset generation complete...')

  print()
  print(repr(builder.info))
  print()

  # Publish data if requested.
  if publish_data_dir is not None:
    logging.info(f'Publishing dataset to {publish_data_dir}')
    _publish_data(
        publish_data_dir=publish_data_dir,
        builder=builder,
        overwrite=args.overwrite,
    )
    logging.info(f'Dataset successfully published to {publish_data_dir}')


def _publish_data_dir(
    publish_dir: Optional[tfds.core.Path], builder: tfds.core.DatasetBuilder
) -> Optional[epath.Path]:
  if publish_dir is None:
    return None
  return epath.Path(publish_dir) / builder.info.full_name


def _publish_data(
    publish_data_dir: epath.Path,
    builder: tfds.core.DatasetBuilder,
    overwrite: bool = False,
) -> None:
  """Publishes the data from the given builder to `publish_data_dir`.

  Arguments:
    publish_data_dir: folder where the data should be published. Should include
      config and version.
    builder: the builder whose data needs to be published.
    overwrite: whether to overwrite existing data in the `publish_root_dir` if
      it exists.
  """
  from_data_dir = epath.Path(builder.data_dir)
  publish_data_dir.mkdir(parents=True, exist_ok=True)
  for filepath in from_data_dir.iterdir():
    filepath.copy(dst=publish_data_dir / filepath.name, overwrite=overwrite)


def _make_download_config(
    args: argparse.Namespace,
    dataset_name: str,
) -> tfds.download.DownloadConfig:
  """Generate the download and prepare configuration."""
  # Load the download config
  manual_dir = args.manual_dir
  if args.add_name_to_manual_dir:
    manual_dir = os.path.join(manual_dir, dataset_name)

  kwargs = {}
  if args.max_shard_size_mb:
    kwargs['max_shard_size'] = args.max_shard_size_mb << 20
  if args.download_config:
    kwargs.update(json.loads(args.download_config))

  if 'download_mode' in kwargs:
    kwargs['download_mode'] = tfds.download.GenerateMode(
        kwargs['download_mode']
    )
  else:
    kwargs['download_mode'] = tfds.download.GenerateMode.REUSE_DATASET_IF_EXISTS
  if args.update_metadata_only:
    kwargs['download_mode'] = tfds.download.GenerateMode.UPDATE_DATASET_INFO

  dl_config = tfds.download.DownloadConfig(
      extract_dir=args.extract_dir,
      manual_dir=manual_dir,
      max_examples_per_split=args.max_examples_per_split,
      register_checksums=args.register_checksums,
      force_checksums_validation=args.force_checksums_validation,
      **kwargs,
  )

  # Add Apache Beam options to download config
  try:
    import apache_beam as beam  # pylint: disable=g-import-not-at-top
  except ImportError:
    beam = None

  if beam is not None:
    if args.beam_pipeline_options:
      dl_config.beam_options = beam.options.pipeline_options.PipelineOptions(
          flags=[f'--{opt}' for opt in args.beam_pipeline_options.split(',')]
      )

  return dl_config


def _get_config_name(
    builder_cls: Type[tfds.core.DatasetBuilder],
    config_kwarg: Optional[str],
    config_name: Optional[str],
    config_idx: Optional[int],
) -> Optional[Union[str, tfds.core.BuilderConfig]]:
  """Extract the config name.

  Args:
    builder_cls: Builder class
    config_kwarg: Config passed as `ds_name/config`
    config_name: Config passed as `--config name`
    config_idx: Config passed as `--config_idx 0`

  Returns:
    The config name to generate, or None.
  """
  num_config = sum(
      c is not None for c in (config_kwarg, config_name, config_idx)
  )
  if num_config > 1:
    raise ValueError(
        'Config should only be defined once by either `ds_name/config`, '
        '`--config` or `--config_idx`'
    )
  elif num_config == 1 and not builder_cls.BUILDER_CONFIGS:
    raise ValueError(
        f'Cannot generate requested config: Dataset {builder_cls.name} does '
        'not have config.'
    )

  if config_kwarg:  # `ds_name/config`
    return config_kwarg
  elif config_name:  # `--config name`
    if config_name.startswith('{'):
      # Json: Dynamically recreate the builder_config
      return builder_cls.builder_config_cls(**json.loads(config_name))
    else:
      return config_name
  elif config_idx is not None:  # `--config_idx 123`
    if config_idx > len(builder_cls.BUILDER_CONFIGS):
      raise ValueError(
          f'--config_idx {config_idx} greater than number '
          f'of configs {len(builder_cls.BUILDER_CONFIGS)} for '
          f'{builder_cls.name}.'
      )
    else:
      # Use `config.name` to avoid
      # https://github.com/tensorflow/datasets/issues/2348
      return builder_cls.BUILDER_CONFIGS[config_idx].name
  else:  # No config defined
    return None



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes and functions to generate the OI Challenge 2019 dataset using Apache Beam."""

import collections
import csv
import io
import json
import os

from absl import logging
import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

beam = tfds.core.lazy_imports.apache_beam
cv2 = tfds.core.lazy_imports.cv2

Metrics = beam.metrics.Metrics


class ReadZipFn(beam.DoFn):
  """Iterates a zip file, yielding filenames and file contents."""

  def process(self, zip_filepath):
    for filename, file in tfds.download.iter_archive(
        zip_filepath, tfds.download.ExtractMethod.ZIP
    ):
      if filename.endswith(".jpg"):
        yield filename, file.read()


class ProcessImageFn(beam.DoFn):
  """Resizes images, re-compresses them in JPEG and yields the result."""

  def __init__(self, target_pixels, jpeg_quality=72):
    self._target_pixels = target_pixels
    self._jpeg_quality = [int(cv2.IMWRITE_JPEG_QUALITY), jpeg_quality]
    self._images_failed = Metrics.counter(self.__class__, "images_failed")

  def process(self, element):
    filename, content = element
    try:
      image = cv2.imdecode(np.frombuffer(content, dtype=np.uint8), flags=3)
    except:
      logging.info("Exception raised while decoding image %s", filename)
      raise
    if image is None:
      self._images_failed.inc()
      logging.info("Image %s could not be decoded", filename)
    else:
      # GIF images contain a single frame.
      if len(image.shape) == 4:  # rank=4 -> rank=3
        image = image.reshape(image.shape[1:])
      # Get image height and width.
      height, width, _ = image.shape
      actual_pixels = height * width
      # If necessary, resize the image to have at most self._target_pixels,
      # keeping the aspect ratio.
      if self._target_pixels and actual_pixels > self._target_pixels:
        factor = np.sqrt(self._target_pixels / actual_pixels)
        image = cv2.resize(image, dsize=None, fx=factor, fy=factor)
      # Encode the image with quality=72 and store it in a BytesIO object.
      _, buff = cv2.imencode(".jpg", image, self._jpeg_quality)
      yield filename, io.BytesIO(buff.tobytes())


class CreateDetectionExampleFn(beam.DoFn):
  """Creates TFDS examples for the Detection track."""

  def __init__(
      self,
      image_labels_filepath,
      box_labels_filepath,
      hierarchy_filepath,
      classes_filepath,
  ):
    self._image_labels_filepath = image_labels_filepath
    self._box_labels_filepath = box_labels_filepath
    self._hierarchy_filepath = hierarchy_filepath
    self._classes_filepath = classes_filepath
    self._image2labels = None
    self._image2boxes = None
    self._hierarchy = None
    self._mid2int = None

  def start(self):
    if self._image_labels_filepath:
      self._image2labels = load_image_level_labels(self._image_labels_filepath)
    if self._box_labels_filepath:
      self._image2boxes = load_box_level_labels(self._box_labels_filepath)
    if self._hierarchy_filepath:
      self._hierarchy = load_class_hierarchy(self._hierarchy_filepath)
    if self._classes_filepath:
      class_descriptions = load_class_descriptions(self._classes_filepath)
      self._mid2int = {mid: i for i, (mid, _) in enumerate(class_descriptions)}

  def process(self, element):
    filename, image_bytes = element
    image_id = os.path.basename(filename).split(".")[0]
    # Image-level annotations.
    objects = []
    if self._image2labels:
      for label, source, confidence in self._image2labels[image_id]:
        objects.append({
            "label": self._mid2int[label],
            "source": source,
            "confidence": confidence,
        })
    # Bounding box-level annotations.
    bobjects = []
    if self._image2boxes:
      for annotation in self._image2boxes[image_id]:
        label, xmin, xmax, ymin, ymax, is_group_of = annotation
        bbox = tfds.features.BBox(xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)
        bobjects.append({
            "label": self._mid2int[label],
            "bbox": bbox,
            "is_group_of": is_group_of,
        })

    yield image_id, {
        "id": image_id,
        "image": image_bytes,
        "objects": objects,
        "bobjects": bobjects,
    }


def load_image_level_labels(filepath):
  """Returns a dictionary mapping image IDs to a list of image-level labels."""
  image2labels = collections.defaultdict(list)
  with tf.io.gfile.GFile(filepath, "r") as csvfile:
    reader = csv.reader(csvfile)
    next(reader)  # Skip header.
    for row in reader:
      if len(row) == 3:
        image_id, label, confidence = row
        source = "verification"
      elif len(row) == 4:
        image_id, source, label, confidence = row
      image2labels[image_id].append((label, source, float(confidence)))
  return image2labels


def load_box_level_labels(filepath):
  """Returns a dictionary mapping image IDs to a list of bounding box annotations."""
  image2boxes = collections.defaultdict(list)
  with tf.io.gfile.GFile(filepath, "r") as csvfile:
    reader = csv.reader(csvfile)
    next(reader)  # Skip header.
    for row in reader:
      if len(row) == 7:
        image_id, label, xmin_s, xmax_s, ymin_s, ymax_s, is_group_of_s = row
      elif len(row) == 13:
        image_id, label = row[0], row[2]
        xmin_s, xmax_s, ymin_s, ymax_s = row[4:8]
        is_group_of_s = row[10]
      xmin, xmax, ymin, ymax = map(float, (xmin_s, xmax_s, ymin_s, ymax_s))
      is_group_of = bool(int(is_group_of_s))
      image2boxes[image_id].append((label, xmin, xmax, ymin, ymax, is_group_of))
  return image2boxes


def load_class_hierarchy(filepath):
  with tf.io.gfile.GFile(filepath, "r") as jsonfile:
    return json.load(jsonfile)


def load_class_descriptions(filepath):
  with tf.io.gfile.GFile(filepath, "r") as csvfile:
    reader = csv.reader(csvfile)
    # Note: this file doesn't have any header.
    return [row for row in reader]


def fill_class_names_in_tfds_info(classes_filepath, tfds_info_features):
  """Fills the class names in ClassLabel features."""
  class_descriptions = load_class_descriptions(classes_filepath)
  mids = [mid for mid, _ in class_descriptions]
  tfds_info_features["objects"]["label"].names = mids
  tfds_info_features["bobjects"]["label"].names = mids



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tree utils."""

import concurrent.futures
import functools
from typing import Callable, TypeVar

from tensorflow_datasets.core.utils import tqdm_utils as tqdm
from tensorflow_datasets.core.utils import type_utils
import tree

Tree = type_utils.Tree
_Tin = TypeVar('_Tin')
_Tout = TypeVar('_Tout')

# Wrapper around `tree.map_structure` to easily switch backend if needed.
map_structure = tree.map_structure


def parallel_map(
    map_fn: Callable[..., _Tout],  # Callable[[_Tin0, _Tin1,...], Tout]
    *trees: Tree[_Tin],  # _Tin0, _Tin1,...
    max_workers: int = 32,
    report_progress: bool = False,
) -> Tree[_Tout]:  # pytype: disable=invalid-annotation
  """Same as `map_structure` but apply map_fn in parallel.

  Args:
    map_fn: Worker function
    *trees: Nested input to pass to the `map_fn`
    max_workers: Number of workers
    report_progress: If True, display a progression bar.

  Returns:
    The nested structure after `map_fn` has been applied.
  """
  with concurrent.futures.ThreadPoolExecutor(
      max_workers=max_workers,
  ) as executor:
    launch_worker = functools.partial(executor.submit, map_fn)
    futures = map_structure(launch_worker, *trees)

    leaves = tree.flatten(futures)

    itr = concurrent.futures.as_completed(leaves)
    if report_progress:
      itr = tqdm.tqdm(itr, total=len(leaves))

    for f in itr:  # Propagate exception to main thread.
      if f.exception():
        raise f.exception()

  return map_structure(lambda f: f.result(), futures)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.features.top_level_feature."""

import time

import numpy as np
from tensorflow_datasets import testing
from tensorflow_datasets.core import features as features_lib
from tensorflow_datasets.core.features import top_level_feature


class FeaturesManagerTest(testing.TestCase):

  def test_sequence_rank(self):
    self.assertEqual(
        1,
        top_level_feature._get_sequence_rank({
            'a': features_lib.TensorInfo(
                shape=(None, 3), dtype=np.int32, sequence_rank=1
            ),
            'b': features_lib.TensorInfo(
                shape=(None,), dtype=np.int32, sequence_rank=1
            ),
        }),
    )

    with self.assertRaisesWithPredicateMatch(
        NotImplementedError, 'mixing sequence and context'
    ):
      top_level_feature._get_sequence_rank({
          'a': features_lib.TensorInfo(shape=(), dtype=np.int32),
          'b': features_lib.TensorInfo(
              shape=(None,), dtype=np.int32, sequence_rank=1
          ),
      })

  def test_flatten_nested(self):
    f = features_lib.FeaturesDict({
        'a': np.int32,
        'b': {
            'c': {
                'd': np.int32,
                'e': np.int32,
            },
        },
        'f': features_lib.Sequence({
            'g': features_lib.Sequence(np.int32),
            'h': np.int32,
        }),
    })

    flat1 = f._flatten({
        'a': 'a',
        'b': {
            'c': {
                'd': {'d': 123},
            },
        },
        'f': {
            'g': 'g',
        },
    })
    self.assertEqual(
        flat1,
        [
            'a',
            {'d': 123},
            None,  # 'e'
            'g',
            None,  # h
        ],
    )
    self.assertEqual(
        f._nest(flat1),
        {
            'a': 'a',
            'b': {
                'c': {
                    'd': {'d': 123},
                    'e': None,
                },
            },
            'f': {
                'g': 'g',
                'h': None,
            },
        },
    )

    f = features_lib.FeaturesDict({
        'a': np.int32,
        'b': {
            'c': np.int32,
        },
    })
    with self.assertRaisesWithPredicateMatch(ValueError, 'received a non dict'):
      f._flatten({'b': 123})

    with self.assertRaisesWithPredicateMatch(
        ValueError, "Unrecognized keys: ['d']"
    ):
      f._flatten({'b': {'c': 123, 'd': 123}})

    with self.assertRaisesWithPredicateMatch(
        ValueError, 'Expected length 2 does not match input length 3'
    ):
      f._nest([None, None, None])

  def test_top_level(self):
    f = features_lib.FeaturesDict({
        'a': np.int32,
        'b': {
            'c': np.int32,
        },
    })

    # Only top level can be decoded
    f.decode_example({
        'a': 1,
        'b': {
            'c': 2,
        },
    })
    f.decode_example_np({
        'a': 1,
        'b': {
            'c': 2,
        },
    })


def test_benchmark_deserialization_speed():
  features = features_lib.FeaturesDict({
      'a': np.int32,
      'b': {
          'c': np.int32,
      },
  })
  example = {'a': 1, 'b': {'c': 2}}
  serialized_example = features.serialize_example(example)
  number_of_examples = 100
  assert features.deserialize_example(serialized_example) == example
  assert features.deserialize_example_np(serialized_example) == example

  start_time = time.time()
  for _ in range(number_of_examples):
    features.deserialize_example(serialized_example)
  tf_time = time.time() - start_time

  start_time = time.time()
  for _ in range(number_of_examples):
    features.deserialize_example_np(serialized_example)
  np_time = time.time() - start_time

  # NumPy deserialization should be faster than TensorFlow deserialization.
  # If you improved TensorFlow deserialization, please delete this test.
  assert np_time < tf_time


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Version utils."""
from __future__ import annotations

import enum
import re
from typing import List, Tuple, Union

from etils import epath

_VERSION_TMPL = r"^(?P<major>{v})" r"\.(?P<minor>{v})" r"\.(?P<patch>{v})$"
_VERSION_WILDCARD_REG = re.compile(_VERSION_TMPL.format(v=r"\d+|\*"))
_VERSION_RESOLVED_REG = re.compile(_VERSION_TMPL.format(v=r"\d+"))


class Experiment(enum.Enum):
  """Experiments which can be enabled/disabled on a per version basis.

  Experiments are designed to gradually apply changes to datasets while
  maintaining backward compatibility with previous versions. All experiments
  should eventually be deleted, once used by all versions of all datasets.

  Eg:
  class Experiment(enum.Enum):
    EXP_A = enum.auto()  # Short description of experiment.

  class MyBuilder(...):
    VERSION = tfds.core.Version('1.2.3', experiments={
        tfds.core.Experiment.EXP_A: True,
        })
  """

  # A Dummy experiment, which should NOT be used, except for testing.
  DUMMY = 1


class Version:
  """Dataset version MAJOR.MINOR.PATCH."""

  _DEFAULT_EXPERIMENTS = {
      Experiment.DUMMY: False,
  }

  def __init__(
      self,
      version: Union[Version, str],
      experiments=None,
      tfds_version_to_prepare=None,
  ):
    """Version init.

    Args:
      version: string. Eg: "1.2.3".
      experiments: dict of experiments. See Experiment.
      tfds_version_to_prepare: string, defaults to None. If set, indicates that
        current version of TFDS cannot be used to `download_and_prepare` the
        dataset, but that TFDS at version {tfds_version_to_prepare} should be
        used instead.
    """
    if isinstance(version, Version):
      version_str = str(version)
      experiments = experiments or version._experiments
      tfds_version_to_prepare = (
          tfds_version_to_prepare or version.tfds_version_to_prepare
      )
    else:
      version_str = version
    self._experiments = self._DEFAULT_EXPERIMENTS.copy()
    self.tfds_version_to_prepare = tfds_version_to_prepare
    if experiments:
      if isinstance(experiments, str):
        raise ValueError(
            f"Invalid Version('{version}', '{experiments}'). Description is "
            "deprecated. RELEASE_NOTES should be used instead."
        )
      self._experiments.update(experiments)
    self.major, self.minor, self.patch = _str_to_version(version_str)

  def implements(self, experiment):
    """Returns True if version implements given experiment."""
    return self._experiments[experiment]

  def __str__(self):
    return "{}.{}.{}".format(*self.tuple)

  def __repr__(self) -> str:
    return f"{type(self).__name__}('{str(self)}')"

  @property
  def tuple(self):
    return self.major, self.minor, self.patch

  def _validate_operand(self, other):
    if isinstance(other, str):
      return Version(other)
    elif isinstance(other, Version):
      return other
    raise AssertionError(
        "{} (type {}) cannot be compared to version.".format(other, type(other))
    )

  def __eq__(self, other):
    other = self._validate_operand(other)
    return self.tuple == other.tuple

  def __ne__(self, other):
    other = self._validate_operand(other)
    return self.tuple != other.tuple

  def __lt__(self, other):
    other = self._validate_operand(other)
    return self.tuple < other.tuple

  def __le__(self, other):
    other = self._validate_operand(other)
    return self.tuple <= other.tuple

  def __gt__(self, other):
    other = self._validate_operand(other)
    return self.tuple > other.tuple

  def __ge__(self, other):
    other = self._validate_operand(other)
    return self.tuple >= other.tuple

  def __hash__(self) -> int:
    return hash(self.tuple)

  def match(self, other_version) -> bool:
    """Returns True if other_version matches.

    Args:
      other_version: string, of the form "x[.y[.x]]" where {x,y,z} can be a
        number or a wildcard.
    """
    major, minor, patch = _str_to_version(other_version, allow_wildcard=True)
    return (
        major in [self.major, "*"]
        and minor in [self.minor, "*"]
        and patch in [self.patch, "*"]
    )

  @classmethod
  def is_valid(cls, version: str) -> bool:
    """Returns True if the version can be parsed."""
    try:
      return cls(version) and True
    except ValueError:  # Invalid version (ex: incomplete data dir)
      return False


def _str_to_version(
    version_str: str, allow_wildcard=False
) -> Tuple[Union[int, str], Union[int, str], Union[int, str]]:
  """Return the tuple (major, minor, patch) version extracted from the str."""
  if not isinstance(version_str, str):
    raise TypeError(
        "Can only convert strings to versions. "
        f"Got: {type(version_str)} with value {version_str}."
    )
  reg = _VERSION_WILDCARD_REG if allow_wildcard else _VERSION_RESOLVED_REG
  res = reg.match(version_str)
  if not res:
    msg = "Invalid version '{}'. Format should be x.y.z".format(version_str)
    if allow_wildcard:
      msg += " with {x,y,z} being digits or wildcard."
    else:
      msg += " with {x,y,z} being digits."
    raise ValueError(msg)
  return tuple(
      v if v == "*" else int(v)  # pylint:disable=g-complex-comprehension
      for v in [res.group("major"), res.group("minor"), res.group("patch")]
  )


def list_all_versions(root_dir: epath.PathLike) -> List[Version]:
  """Lists all dataset versions present on disk, sorted."""
  root_dir = epath.Path(root_dir)
  versions = []
  try:
    for path in root_dir.iterdir():
      if Version.is_valid(path.name) and path.is_dir():
        versions.append(Version(path.name))
  except OSError:
    return versions
  return sorted(versions)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for CivilComments from Jigsaw Unintended Bias Kaggle Competition."""

from tensorflow_datasets import testing
from tensorflow_datasets.text import civil_comments


class CivilCommentsNoCovertTest(testing.DatasetBuilderTestCase):
  DATASET_CLASS = civil_comments.CivilComments
  BUILDER_CONFIG_NAMES_TO_TEST = [
      "CivilComments",
      "CivilCommentsIdentities",
  ]
  SPLITS = {
      "train": 3,  # Number of fake train examples
      "test": 1,  # Number of fake test examples
      "validation": 1,  # Number of fake validation examples
  }


class CivilCommentsToxicSpans(testing.DatasetBuilderTestCase):
  DATASET_CLASS = civil_comments.CivilComments
  BUILDER_CONFIG_NAMES_TO_TEST = ["CivilCommentsToxicSpans"]
  SPLITS = {
      "train": 2,  # Number of fake train examples
      "test": 1,  # Number of fake test examples
      "validation": 1,  # Number of fake validation examples
  }


class CivilCommentsInContext(testing.DatasetBuilderTestCase):
  DATASET_CLASS = civil_comments.CivilComments
  BUILDER_CONFIG_NAMES_TO_TEST = ["CivilCommentsInContext"]
  SPLITS = {
      "train": 2,  # Number of fake train examples
  }


# Separate test for CivilCommentsCovert without validation split.
class CivilCommentsCovertTest(testing.DatasetBuilderTestCase):
  DATASET_CLASS = civil_comments.CivilComments
  BUILDER_CONFIG_NAMES_TO_TEST = ["CivilCommentsCovert"]
  SPLITS = {
      "train": 3,  # Number of fake train examples
      "test": 1,  # Number of fake test examples
  }


if __name__ == "__main__":
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module defines the methods a logger implementation should define."""

from typing import Any, Dict, Optional, Union

from tensorflow_datasets.core import decode
from tensorflow_datasets.core import download as download_lib
from tensorflow_datasets.core import splits as splits_lib
from tensorflow_datasets.core.file_adapters import FileFormat
from tensorflow_datasets.core.logging import call_metadata
from tensorflow_datasets.core.utils import read_config as read_config_lib
from tensorflow_datasets.core.utils import type_utils

TreeDict = type_utils.TreeDict


class Logger:
  """Defines interface any TFDS logger must implement.

  Registered loggers methods are called on TFDS events, synchronously, from the
  same thread the call was made, in sequence, and in registration order.
  Exceptions are *NOT* caught.
  """

  def tfds_import(
      self,
      *,
      metadata: call_metadata.CallMetadata,
      import_time_ms_tensorflow: int,
      import_time_ms_dataset_builders: int,
  ):
    """Callback called when user calls `import tensorflow_datasets`."""
    pass

  def builder_init(
      self,
      *,
      metadata: call_metadata.CallMetadata,
      name: str,
      data_dir: Optional[str],
      config: Optional[str],
      version: Optional[str],
  ):
    """Callback called when user calls `DatasetBuilder(...)`."""
    pass

  def builder_info(
      self,
      *,
      metadata: call_metadata.CallMetadata,
      name: str,
      config_name: Optional[str],
      version: str,
      data_path: str,
  ):
    """Callback called when user calls `builder.info()`."""
    pass

  def as_dataset(
      self,
      *,
      metadata: call_metadata.CallMetadata,
      name: str,
      config_name: Optional[str],
      version: str,
      data_path: str,
      split: Optional[type_utils.Tree[splits_lib.SplitArg]],
      batch_size: Optional[int],
      shuffle_files: bool,
      read_config: read_config_lib.ReadConfig,
      as_supervised: bool,
      decoders: Optional[TreeDict[decode.partial_decode.DecoderArg]],
  ):
    """Callback called when user calls `dataset_builder.as_dataset`.

    Callback is also triggered by `tfds.load`, which calls `as_dataset`.
    The logger MUST NOT mutate passed objects (decoders, read_config, ...).

    Args:
      metadata: CallMetadata associated to call.
      name: the name of the dataset. E.g.: "mnist".
      config_name: the name of the config or None.
      version: the dataset version. E.g.: "1.2.3".
      data_path: The path to directory of the dataset loaded. E.g.:
        "/home/alice/.tensorflow_datasets/mnist/1.2.3".
      split: name of the split requested by user. 'all' for all splits.
      batch_size: See DatasetBuilder.as_dataset docstring.
      shuffle_files: -
      read_config: -
      as_supervised: -
      decoders: flatten dict of decoders dict given to `as_dataset`, with the
        values being `{module_name}.{class_name}` instad of a
    """
    pass

  def download_and_prepare(
      self,
      *,
      metadata: call_metadata.CallMetadata,
      name: str,
      config_name: Optional[str],
      version: str,
      data_path: str,
      download_dir: Optional[str],
      download_config: Optional[download_lib.DownloadConfig],
      file_format: Union[None, str, FileFormat],
  ):
    """Callback called when user calls `dataset_builder.download_and_prepare`."""
    pass

  def as_numpy(self, *, metadata: call_metadata.CallMetadata, dataset: Any):
    """Callback called when user calls `tfds.as_numpy(...)`."""
    pass

  def builder(
      self,
      *,
      metadata: call_metadata.CallMetadata,
      name: str,
      try_gcs: Optional[bool],
  ):
    """Callback called when user calls `tfds.builder(...)`."""
    pass

  def dataset_collection(
      self,
      metadata: call_metadata.CallMetadata,
      name: str,
      loader_kwargs: Optional[Dict[str, Any]],
  ):
    """Callback called when user calls `tfds.dataset_collection(...)`."""
    pass

  def load(
      self,
      *,
      metadata: call_metadata.CallMetadata,
      name: str,
      split: Optional[type_utils.Tree[splits_lib.SplitArg]],
      data_dir: Optional[str],
      batch_size: Optional[int],
      shuffle_files: Optional[bool],
      download: Optional[bool],
      as_supervised: Optional[bool],
      decoders: Optional[TreeDict[decode.partial_decode.DecoderArg]],
      read_config: Optional[read_config_lib.ReadConfig],
      with_info: Optional[bool],
      try_gcs: Optional[bool],
  ):
    """Callback called when user calls `tfds.load(...)`."""
    pass

  def list_builders(
      self,
      *,
      metadata: call_metadata.CallMetadata,
      with_community_datasets: Optional[bool],
  ):
    """Callback called when user calls `tfds.list_builders(...)`."""
    pass

  def process_ends(self):
    """Called when the process is about to end (atexit)."""
    pass

  def data_source(
      self,
      *,
      metadata: call_metadata.CallMetadata,
      name: str,
      split: Optional[type_utils.Tree[splits_lib.SplitArg]],
      data_dir: Optional[str],
      download: Optional[bool],
      decoders: Optional[TreeDict[decode.partial_decode.DecoderArg]],
      try_gcs: Optional[bool],
  ):
    """Callback called when user calls `tfds.data_source(...)`."""
    pass

  def as_data_source(
      self,
      *,
      metadata: call_metadata.CallMetadata,
      name: str,
      config_name: Optional[str],
      version: str,
      data_path: str,
      split: Optional[type_utils.Tree[splits_lib.SplitArg]],
      decoders: Optional[TreeDict[decode.partial_decode.DecoderArg]],
  ):
    """Callback called when user calls `dataset_builder.as_data_source(...)`."""
    pass



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Custom Translate Datasets Template."""

from __future__ import annotations

import collections
import os
from typing import Dict, List, Tuple

from etils import epath
from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import dataset_info
from tensorflow_datasets.core import features as features_lib
from tensorflow_datasets.core import splits as split_lib
from tensorflow_datasets.core.utils import version
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

# Dict of 'split_name'-> 'language' -> `List[text_data]`
SplitExampleDict = Dict[str, Dict[str, List[str]]]


class TranslateFolder(dataset_builder.DatasetBuilder):
  """Generic text translation dataset created from manual directory.

  The directory content should be as followed:

  ```
  path/to/my_data/
    lang1.train.txt
    lang2.train.txt
    lang1.test.txt
    lang2.test.txt
    ...
  ```

  Each files should have one example per line. Line order should match between
  files.

  To use it:

  ```
  builder = tfds.TranslateFolder(root_dir='path/to/my_data/')
  print(builder.info)  # Splits, num examples,... are automatically calculated
  ds = builder.as_dataset(split='train', shuffle_files=True)
  ```

  Note: All examples from all splits are loaded in memory in `__init__`.
  """

  VERSION = version.Version('1.0.0')

  def __init__(self, root_dir: str):
    # Extract the splits, examples
    root_dir = os.path.expanduser(root_dir)
    self._split_examples, self._languages = _get_split_language_examples(
        root_dir
    )

    super(TranslateFolder, self).__init__()
    # Reset `_data_dir` as it should not change to DATA_DIR/Version
    self._data_dir = root_dir

    # Update DatasetInfo splits
    split_infos = [
        split_lib.SplitInfo(  # pylint: disable=g-complex-comprehension
            name=split_name,
            shard_lengths=[len(next(iter(examples.values())))],
            num_bytes=0,
        )
        for split_name, examples in self._split_examples.items()
    ]
    split_dict = split_lib.SplitDict(split_infos)
    self.info.set_splits(split_dict)

  def _info(self) -> dataset_info.DatasetInfo:
    return dataset_info.DatasetInfo(
        builder=self,
        description='Generic text translation dataset.',
        features=features_lib.FeaturesDict(
            {lang: features_lib.Text() for lang in self._languages}
        ),
    )

  def _download_and_prepare(self, **kwargs):  # pytype: disable=signature-mismatch  # overriding-parameter-count-checks
    raise NotImplementedError(
        'No need to call download_and_prepare function for {}.'.format(
            type(self).__name__
        )
    )

  def download_and_prepare(self, **kwargs):
    return self._download_and_prepare()

  def _as_dataset(
      self, split, shuffle_files=False, decoders=None, read_config=None
  ) -> tf.data.Dataset:
    """Generate dataset for given split."""
    del read_config  # Unused (automatically created in `DatasetBuilder`)
    if decoders:
      raise NotImplementedError(
          '`decoders` is not supported with {}'.format(type(self).__name__)
      )
    if split not in self.info.splits.keys():
      raise ValueError(
          'Unrecognized split {}. Subsplit API not yet supported for {}. '
          'Split name should be one of {}.'.format(
              split, type(self).__name__, list(self.info.splits.keys())
          )
      )

    # Build the tf.data.Dataset object
    lang_example_dict = self._split_examples[split]
    ds = tf.data.Dataset.from_tensor_slices(lang_example_dict)
    if shuffle_files:
      ds = ds.shuffle(len(lang_example_dict))
    return ds


def _get_split_language_examples(
    root_dir: str,
) -> Tuple[SplitExampleDict, List[str]]:
  """Extract all split names and associated text data.

  Args:
    root_dir: The folder where the `lang.split.txt` are located

  Returns:
    split_examples: Mapping split_names -> language -> List[text_data]
    languages: The list of languages
  """
  split_examples = collections.defaultdict(dict)
  languages = set()
  files = tf.io.gfile.listdir(root_dir)
  for file in files:
    lang, split_name, _ = file.split('.')
    split_examples[split_name][lang] = _list_examples(
        os.path.join(root_dir, file)
    )
    languages.add(lang)

  # One-to-One translation
  for split, examples in split_examples.items():
    num_examples = {lang: len(ex) for lang, ex in examples.items()}
    if len(set(num_examples.values())) != 1:
      raise ValueError(
          'Num examples for split {} do not match: {}'.format(
              split, num_examples
          )
      )

  return split_examples, sorted(languages)


def _list_examples(file: str) -> List[str]:
  with epath.Path(file).open() as f:
    sentences = f.read().splitlines()
  return sentences



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.community.register_path."""

from etils import epath
from tensorflow_datasets import testing
from tensorflow_datasets.core import naming
from tensorflow_datasets.core.community import register_path


def test_data_dir_register():
  register = register_path.DataDirRegister(
      namespace_to_data_dirs={'ns1': [epath.Path('/path/ns1')]}
  )
  assert {'ns1'} == register.namespaces


def test_list_dataset_references(mock_fs: testing.MockFs):
  path = '/path/ns1'
  mock_fs.add_file(path=f'{path}/ds1/1.0.0/features.json')
  mock_fs.add_file(path=f'{path}/ds2/config1/1.0.0/features.json')
  mock_fs.add_file(path=f'{path}/ds2/config1/2.0.0/features.json')
  mock_fs.add_file(path=f'{path}/ds2/config2/1.0.0/features.json')
  data_dir = epath.Path('/path/ns1')
  register = register_path.DataDirRegister(
      namespace_to_data_dirs={'ns1': [data_dir]}
  )
  assert sorted(register.list_dataset_references()) == [
      naming.DatasetReference(
          dataset_name='ds1', namespace='ns1', data_dir=data_dir
      ),
      naming.DatasetReference(
          dataset_name='ds2', namespace='ns1', data_dir=data_dir
      ),
  ]



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""i_naturalist2021 dataset."""

from tensorflow_datasets.image_classification.i_naturalist2021 import i_naturalist2021
import tensorflow_datasets.public_api as tfds


class INaturalist2021Test(tfds.testing.DatasetBuilderTestCase):
  """Tests for i_naturalist2021 dataset."""

  DATASET_CLASS = i_naturalist2021.INaturalist2021
  SPLITS = {
      "mini": 2,  # Number of fake mini examples
      "test": 3,  # Number of fake test examples
      "train": 3,  # Number of fake train examples
      "val": 2,  # Number of fake val examples
  }
  OVERLAPPING_SPLITS = ["mini", "train"]

  DL_EXTRACT_RESULT = {}
  for split, split_file in i_naturalist2021._SPLIT_FILENAMES.items():
    DL_EXTRACT_RESULT[f"{split}_img"] = f"{split_file}.tar.gz"
    DL_EXTRACT_RESULT[f"{split}_json"] = split_file

  SKIP_CHECKSUMS = True


if __name__ == "__main__":
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Transformation library.

Note that this is an experimental new feature, so the API may change.
"""

from __future__ import annotations

import functools
from typing import Any, Callable, Iterator, List, Mapping, Optional, Union

import numpy as np

from tensorflow_datasets.core import split_builder as split_builder_lib

Key = split_builder_lib.Key
Example = split_builder_lib.Example
KeyExample = split_builder_lib.KeyExample
ExampleTransformFn = Callable[[Example], Iterator[Example]]


def remove_feature(feature_name: Union[str, List[str]]) -> ExampleTransformFn:
  """Removes the feature(s) with the given name."""

  if isinstance(feature_name, str):
    feature_name = [feature_name]

  def apply_on_example(example: Example) -> Iterator[Example]:
    for name in feature_name:
      del example[name]
    yield example

  return apply_on_example


def rename_feature(from_: str, to: str) -> ExampleTransformFn:
  """Renames the feature.

  Note that if `to` already exists in the example, then it will be overridden.

  Arguments:
    from_: the name of the feature that must be renamed.
    to: the feature name to which it must be renamed.

  Returns:
    function that renames the feature of an example.
  """

  def apply_on_example(example: Example) -> Iterator[Example]:
    example[to] = example.pop(from_)
    yield example

  return apply_on_example


def rename_features(name_mapping: Mapping[str, str]) -> ExampleTransformFn:
  """Renames all the feature specified in the given mapping.

  Note that if the 'to feature name' already exists in the example, then it will
  be overridden. The order in which the renames are doing is not guaranteed, so
  this should not be used to swap the content of two features.

  Arguments:
    name_mapping: mapping from the feature name that must be renamed to the new
      feature name.

  Returns:
    function that renames the feature of an example.
  """

  def apply_on_example(example: Example) -> Iterator[Example]:
    for from_name, to_name in name_mapping.items():
      example[to_name] = example.pop(from_name)
    yield example

  return apply_on_example


def _transform_example(
    example: Example,
    fn: Callable[[Any], Any],
    in_parts: List[str],
    out_parts: List[str],
) -> Example:
  """Transforms the specified input entry with `fn`.

  Note that the given example data is transformed in place.

  Arguments:
    example: the example data that needs to be transformed.
    fn: the function used to transform data.
    in_parts: what feature in the example should be transformed. For example, if
      `in_parts` is `["a", "b"]`, then `fn` will be applied to the nested
      feature `b` inside `a`.
    out_parts: where the transformed data should be stored in the example. Note
      that this needs to be at the same level as `in_parts`. For example, if
      `in_parts=["a", "b", "c"]`, then the output feature needs to be inside
      `["a", "b']`, e.g. `out_parts=["a", "b", "new_feature"]`.

  Returns:
    the transformed example.
  """
  if not in_parts:
    raise ValueError("Should be called with non-empty `in_parts`!")
  in_key = in_parts[0]
  out_key = out_parts[0]
  if len(in_parts) > 1:
    in_rest = in_parts[1:]
    out_rest = out_parts[1:]
    if isinstance(example[in_key], np.ndarray) or isinstance(
        example[in_key], list
    ):
      example[out_key] = [
          _transform_example(
              example=ex, fn=fn, in_parts=in_rest, out_parts=out_rest
          )
          for ex in example[in_key]
      ]
    else:
      example[out_key] = _transform_example(
          example=example[in_key], fn=fn, in_parts=in_rest, out_parts=out_rest
      )
  elif len(in_parts) == 1:
    input_data = example[in_key]
    if isinstance(input_data, List):
      example[out_key] = [fn(x) for x in input_data]
    else:
      example[out_key] = fn(example[in_key])
  return example


def apply_fn(
    fn: Callable[[Any], Any],
    input_feature: str,
    output_feature: Optional[str] = None,
) -> ExampleTransformFn:
  """Returns a function that applies the given `fn` on the `input_feature`.

  Arguments:
    fn: the function to apply.
    input_feature: the name of the feature on which `fn` should be applied. If a
      nested feature needs to be transformed, then specify it using slashes. For
      example, to transform feature `b` inside `a`, use `a/b`.
    output_feature: the name of the feature where the result should be stored.
      If `None`, then `input_feature` will be overridden.

  Returns:
    a function that applies `fn` to `input_feature` field of examples and stores
    the result in `output_feature`.
  """
  input_parts = input_feature.split("/")
  if output_feature is not None:
    output_parts = output_feature.split("/")
  else:
    output_parts = input_parts
  if (
      len(input_parts) != len(output_parts)
      or input_parts[:-1] != output_parts[:-1]
  ):
    raise ValueError(
        "The out-feature must have the same ancestor as the "
        f"in-feature! Got in={input_parts}, out={output_parts}."
    )

  def transform(example: Example) -> Iterator[Example]:
    yield _transform_example(
        example=example, fn=fn, in_parts=input_parts, out_parts=output_parts
    )

  return transform


def _apply_filter(
    example: Example,
    fn: Callable[[Any], Any],
    in_parts: List[str],
) -> Iterator[Example]:
  in_data = example
  for part in in_parts:
    in_data = in_data[part]
  if fn(in_data):
    yield example


def apply_filter(
    fn: Callable[[Any], bool],
    input_feature: Optional[str] = None,
) -> ExampleTransformFn:
  """Filters examples for whom `fn` returns true on `input_feature`.

  Arguments:
    fn: the function that returns whether the input feature of an example should
      be filtered.
    input_feature: the input feature to apply `fn` to. If this is the empty
      string, then `fn` will receive the entire example.

  Returns:
    function that can be applied to filter examples.
  """
  if input_feature:
    input_parts = input_feature.split("/")
  else:
    input_parts = []
  return functools.partial(_apply_filter, fn=fn, in_parts=input_parts)


def _apply_do_fn(
    example: Example,
    fn: Callable[[Any], Iterator[Any]],
) -> Iterator[Example]:
  yield from fn(example)


def apply_do_fn(
    fn: Callable[[Any], Iterator[Any]],
) -> ExampleTransformFn:
  """Applies `fn` that can return any number of examples based on an input.

  A do_fn can be used for multiple purposes such as augmentations, but also to
  filter examples.

  Arguments:
    fn: the function that takes an example and outputs 0 or more examples.

  Returns:
    function that can be applied to perform a do fn.
  """
  return functools.partial(_apply_do_fn, fn=fn)


def _apply_transformations_no_key(
    example: Example,
    transformations: List[ExampleTransformFn],
) -> Iterator[Example]:
  """Transforms the example with the given transformations."""
  if transformations:
    fn = transformations[0]
    other_fns = transformations[1:]
    for v in fn(example):
      yield from _apply_transformations_no_key(
          example=v, transformations=other_fns
      )
  else:
    yield example


def apply_transformations(
    key: Key,
    example: Example,
    transformations: List[ExampleTransformFn],
) -> Iterator[KeyExample]:
  for i, transformed_example in enumerate(
      _apply_transformations_no_key(
          example=example, transformations=transformations
      )
  ):
    yield f"{key}_{i}", transformed_example



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DatasetCollection base class."""
import dataclasses
import inspect
from typing import Any, List, Mapping, Optional, Type, Union

from etils import epath
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import registered
from tensorflow_datasets.core.utils import version as version_lib

DESCRIPTION_FILE = "description.md"
CITATIONS_FILE = "citations.bib"


def get_filepath_in_dataset_folder(
    dataset_cls: Type[Any], file_name: str
) -> epath.Path:
  directory_path = epath.Path(inspect.getfile(dataset_cls)).parent
  return directory_path / file_name


def get_file_content_from_dataset_folder(
    dataset_class: Type[Any],
    file_name: str,
    raise_error_if_fails: bool = False,
) -> Optional[str]:
  """Returns the file content from the relevant dataset folder.

  Args:
    dataset_class: The dataset collection class for which the file has to be
      read.
    file_name: The name of the file to load.
    raise_error_if_fails: Whether to raise an error if the file's content cannot
      be retrieved.

  Returns:
    The requested file's content. If unavailable, it will return None if
    raise_error_if_fails is set to False, otherwise it will raise the
    encountered error.
  """
  file_path = get_filepath_in_dataset_folder(dataset_class, file_name)
  try:
    return file_path.read_text()
  except Exception as e:  # pylint: disable=broad-except
    if raise_error_if_fails:
      raise e
    else:
      return None


@dataclasses.dataclass
class DatasetCollectionInfo:
  """Information about a dataset collection.

  `DatasetCollectionInfo` documents a dataset collection, including its name,
  description, release notes and citations.

  Attributes:
    name: The name of the dataset collection.
    description: A markdown-formatted description of the dataset collection.
    release_notes: A mapping of dataset collection's versions with their
      corresponding release notes.
    citation: Optional citation for the dataset collection.
    homepage: Optional homepage for the dataset collection.
  """

  name: str
  description: str
  release_notes: Mapping[str, str]
  citation: Optional[str] = None
  homepage: Optional[str] = None

  @classmethod
  def from_cls(
      cls,
      dataset_collection_class: Type["DatasetCollection"],
      release_notes: Mapping[str, str],
      description: Optional[str] = None,
      citation: Optional[str] = None,
      homepage: Optional[str] = None,
  ) -> "DatasetCollectionInfo":
    """Creates a DatasetCollectionInfo instance based on class information."""
    name: str = naming.camelcase_to_snakecase(dataset_collection_class.__name__)
    if not description:
      description = get_file_content_from_dataset_folder(
          dataset_collection_class, DESCRIPTION_FILE, raise_error_if_fails=True
      )
    if not citation:
      citation = get_file_content_from_dataset_folder(
          dataset_collection_class, CITATIONS_FILE
      )
    return cls(
        name=name,
        release_notes=release_notes,
        description=description,
        citation=citation,
        homepage=homepage,
    )


class DatasetCollection(
    registered.RegisteredDatasetCollection, skip_registration=True
):
  """Base class to define a dataset collection.

  Subclasses should overwrite `info` to return populated DatasetCollectionInfo.

  Subclasses should also overwrite `datasets` to return a dictionary of versions
  to the datasets included in that collection's version.
  """

  @property
  def info(self) -> DatasetCollectionInfo:
    raise NotImplementedError

  @property
  def datasets(self) -> Mapping[str, Mapping[str, naming.DatasetReference]]:
    """Returns the datasets included in the collection, ordered by version.

    Users will need to overwrite this function when implementing their dataset
    collection.

    The returned dictionary needs to contain the dataset collection versions as
    keys, and a dictionary of the included TFDS datasets as values.

    For example:
    ```
    @property
    def datasets(self):
    return {
        "1.0.0":
            naming.references_for({
                "yes_no": "yes_no:1.0.0",
                "sst2": "glue/sst:2.0.0",
                "assin2": "assin2:1.0.0",
            }),
        ...
    }
    ```

    Note that the above is equivalent to:
    ```
    @property
    def datasets(self):
    return {
        "1.0.0": {
            "yes_no":
                naming.DatasetReference(
                    dataset_name="yes_no", version="1.0.0"),
            "sst2":
                naming.DatasetReference(
                    dataset_name="glue", config="sst2", version="2.0.0"),
            "assin2":
                naming.DatasetReference(
                    dataset_name="assin2", version="1.0.0"),
        },
        ...
    }
    ```
    """
    raise NotImplementedError

  def __repr__(self):
    return f"DatasetCollection(info={self.info}, datasets={self.datasets})"

  @property
  def all_versions(self) -> List[version_lib.Version]:
    """Returns all versions available for the dataset collection."""
    return [
        version_lib.Version(version_str) for version_str in self.datasets.keys()
    ]

  def get_latest_version(self) -> str:
    """Returns the latest version of this dataset collection."""
    return str(max(self.all_versions))

  def get_collection(
      self,
      version: Union[None, str, version_lib.Version] = None,
  ) -> Mapping[str, naming.DatasetReference]:
    """Returns the requested versioned dataset collection.

    Args:
      version: The requested version. If no version is specified, returns the
        most recently added version.

    Returns:
      The requested dataset collection.
    """
    if not version:
      return self.datasets[self.get_latest_version()]

    if isinstance(version, version_lib.Version):
      version = str(version)
    for v in reversed(self.all_versions):
      if v.match(version):
        return self.datasets[str(v)]

    raise ValueError(f"No datasets could be retrieved for version {version}")

  def list_datasets(
      self,
      version: Union[None, str, version_lib.Version] = None,
  ) -> str:
    """Returns the datasets included in a versioned dataset collection."""
    msgs = [f"The dataset collection {self.info.name}"]
    if version:
      msgs.append(f"(version: {version})")
    msgs.append("contains the datasets:\n")
    versioned_collection = self.get_collection(version)
    for ds_name, benchmark_spec in versioned_collection.items():
      msgs.append(f"- {ds_name}: {benchmark_spec}\n")
    return " ".join(msgs)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""UCF-101 dataset from https://www.crcv.ucf.edu/data/UCF101.php."""

import os

from absl import logging
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

UCF_101_URL = 'https://storage.googleapis.com/thumos14_files/UCF101_videos.zip'
SPLITS_URL = (
    'https://www.crcv.ucf.edu/data/UCF101/'
    'UCF101TrainTestSplits-RecognitionTask.zip'
)

_CITATION = """\
@article{DBLP:journals/corr/abs-1212-0402,
  author    = {Khurram Soomro and
               Amir Roshan Zamir and
               Mubarak Shah},
  title     = {{UCF101:} {A} Dataset of 101 Human Actions Classes From Videos in
               The Wild},
  journal   = {CoRR},
  volume    = {abs/1212.0402},
  year      = {2012},
  url       = {http://arxiv.org/abs/1212.0402},
  archivePrefix = {arXiv},
  eprint    = {1212.0402},
  timestamp = {Mon, 13 Aug 2018 16:47:45 +0200},
  biburl    = {https://dblp.org/rec/bib/journals/corr/abs-1212-0402},
  bibsource = {dblp computer science bibliography, https://dblp.org}
}
"""

_LABELS_FNAME = 'video/ucf101_labels.txt'


class Ucf101Config(tfds.core.BuilderConfig):
  """ "Configuration for UCF101 split and possible video rescaling."""

  def __init__(self, *, split_number, width=None, height=None, **kwargs):
    """The parameters specifying how the dataset will be processed.

    The dataset comes with three separate splits. You can specify which split
    you want in `split_number`. If `width` and `height` are set, the videos
    will be rescaled to have those heights and widths (using ffmpeg).

    Args:
      split_number: The split number, one of (1, 2, 3)
      width: An integer with the width or None.
      height: An integer with the height or None.
      **kwargs: Passed on to the constructor of `BuilderConfig`.
    """
    super(Ucf101Config, self).__init__(
        version=tfds.core.Version('2.0.0'),
        release_notes={
            '2.0.0': 'New split API (https://tensorflow.org/datasets/splits)',
        },
        **kwargs,
    )
    if (width is None) ^ (height is None):
      raise ValueError('Either both dimensions should be set, or none of them')
    self.width = width
    self.height = height
    if split_number not in (1, 2, 3):
      raise ValueError(
          'Unknown split number {}, should be 1, 2 or 3'.format(split_number)
      )
    self.split_number = split_number


class Ucf101(tfds.core.GeneratorBasedBuilder):
  """Ucf101 action recognition dataset.

  Note that in contrast to the labels provided in the original dataset, here the
  labels start at zero, not at one.
  """

  BUILDER_CONFIGS = [
      Ucf101Config(
          name='ucf101_1_256',
          description='256x256 UCF with the first action recognition split.',
          width=256,
          height=256,
          split_number=1,
      ),
      Ucf101Config(
          name='ucf101_1',
          description='UCF with the action recognition split #1.',
          width=None,
          height=None,
          split_number=1,
      ),
      Ucf101Config(
          name='ucf101_2',
          description='UCF with the action recognition split #2.',
          width=None,
          height=None,
          split_number=2,
      ),
      Ucf101Config(
          name='ucf101_3',
          description='UCF with the action recognition split #3.',
          width=None,
          height=None,
          split_number=3,
      ),
  ]

  def _info(self):
    if self.builder_config.width is not None:
      if self.builder_config.height is None:
        raise ValueError('Provide either both height and width or none.')
      ffmpeg_extra_args = (
          '-vf',
          'scale={}x{}'.format(
              self.builder_config.height, self.builder_config.width
          ),
      )
    else:
      ffmpeg_extra_args = []

    video_shape = (
        None,
        self.builder_config.height,
        self.builder_config.width,
        3,
    )
    labels_names_file = tfds.core.tfds_path(_LABELS_FNAME)
    features = tfds.features.FeaturesDict({
        'video': tfds.features.Video(
            video_shape,
            ffmpeg_extra_args=ffmpeg_extra_args,
            encoding_format='jpeg',
        ),  # pytype: disable=wrong-arg-types  # gen-stub-imports
        'label': tfds.features.ClassLabel(names_file=labels_names_file),
    })
    return tfds.core.DatasetInfo(
        builder=self,
        description='A 101-label video classification dataset.',
        features=features,
        homepage='https://www.crcv.ucf.edu/data-sets/ucf101/',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    splits_folder = 'ucfTrainTestlist'

    urls_to_download = {
        'videos': UCF_101_URL,
        'splits': SPLITS_URL,
    }
    downloaded_urls = dl_manager.download_and_extract(urls_to_download)

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                'videos_dir': downloaded_urls['videos'],
                'splits_dir': downloaded_urls['splits'],
                'data_list': '{}/trainlist{:02d}.txt'.format(
                    splits_folder, self.builder_config.split_number
                ),
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={
                'videos_dir': downloaded_urls['videos'],
                'splits_dir': downloaded_urls['splits'],
                'data_list': '{}/testlist{:02d}.txt'.format(
                    splits_folder, self.builder_config.split_number
                ),
            },
        ),
    ]

  def _generate_examples(self, videos_dir, splits_dir, data_list):
    data_list_path_path = os.path.join(splits_dir, data_list)
    with tf.io.gfile.GFile(data_list_path_path, 'r') as data_list_file:
      labels_and_paths = data_list_file.readlines()
    for label_and_path in sorted(labels_and_paths):
      # The train splits contain not only the filename, but also a digit
      # encoding the label separated by a space, which we ignore.
      label_and_path = label_and_path.strip().split(' ')[0]
      label, path = os.path.split(label_and_path)
      # Fix an inconsistency between the names in the list and in the zip file.
      path = path.replace('HandStandPushups', 'HandstandPushups')
      video_path = os.path.join(videos_dir, 'UCF101', path)
      if not tf.io.gfile.exists(video_path):
        logging.error('Example %s not found', video_path)
        continue
      # We extract the label from the filename.
      yield path, {'video': video_path, 'label': label}



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Long T5 dataset collection."""
import collections

from typing import Mapping

from tensorflow_datasets.core import dataset_collection_builder
from tensorflow_datasets.core import naming


class Longt5(dataset_collection_builder.DatasetCollection):
  """Long T5 dataset collection."""

  @property
  def info(self) -> dataset_collection_builder.DatasetCollectionInfo:
    return dataset_collection_builder.DatasetCollectionInfo.from_cls(
        dataset_collection_class=self.__class__,
        release_notes={
            "1.0.0": "Initial release",
        },
        homepage="https://github.com/google-research/longt5",
    )

  @property
  def datasets(
      self,
  ) -> Mapping[str, Mapping[str, naming.DatasetReference]]:
    return collections.OrderedDict(
        {
            "1.0.0": naming.references_for({
                "natural_questions": "natural_questions/longt5:0.1.0",
                "media_sum": "media_sum:1.0.0",
            })
        }
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The Definite Pronoun Resolution Dataset."""

from etils import epath
import tensorflow_datasets.public_api as tfds

_CITATION = """\
@inproceedings{rahman2012resolving,
  title={Resolving complex cases of definite pronouns: the winograd schema challenge},
  author={Rahman, Altaf and Ng, Vincent},
  booktitle={Proceedings of the 2012 Joint Conference on Empirical Methods in Natural Language Processing and Computational Natural Language Learning},
  pages={777--789},
  year={2012},
  organization={Association for Computational Linguistics}
}"""

_DESCRIPTION = """\
Composed by 30 students from one of the author's undergraduate classes. These
sentence pairs cover topics ranging from real events (e.g., Iran's plan to
attack the Saudi ambassador to the U.S.) to events/characters in movies (e.g.,
Batman) and purely imaginary situations, largely reflecting the pop culture as
perceived by the American kids born in the early 90s. Each annotated example
spans four lines: the first line contains the sentence, the second line contains
the target pronoun, the third line contains the two candidate antecedents, and
the fourth line contains the correct antecedent. If the target pronoun appears
more than once in the sentence, its first occurrence is the one to be resolved.
"""

_DATA_URL_PATTERN = 'https://s3.amazonaws.com/datasets.huggingface.co/definite_pronoun_resolution/{}.c.txt'


class DefinitePronounResolution(tfds.core.GeneratorBasedBuilder):
  """The Definite Pronoun Resolution Dataset."""

  VERSION = tfds.core.Version('1.1.0')

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            'sentence': tfds.features.Text(),
            'pronoun': tfds.features.Text(),
            'candidates': tfds.features.Sequence(
                tfds.features.Text(), length=2
            ),
            'label': tfds.features.ClassLabel(num_classes=2),
        }),
        supervised_keys=('sentence', 'label'),
        homepage='https://www.hlt.utdallas.edu/~vince/data/emnlp12/',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    files = dl_manager.download({
        'train': _DATA_URL_PATTERN.format('train'),
        'test': _DATA_URL_PATTERN.format('test'),
    })
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST, gen_kwargs={'filepath': files['test']}
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN, gen_kwargs={'filepath': files['train']}
        ),
    ]

  def _generate_examples(self, filepath):
    with epath.Path(filepath).open() as f:
      line_num = -1
      while True:
        line_num += 1
        sentence = f.readline().strip()
        pronoun = f.readline().strip()
        candidates = [c.strip() for c in f.readline().strip().split(',')]
        correct = f.readline().strip()
        f.readline()
        if not sentence:
          break
        yield line_num, {
            'sentence': sentence,
            'pronoun': pronoun,
            'candidates': candidates,
            'label': candidates.index(correct),
        }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""A refactoring script to refactor the dataset files into a single folder.

```
my_dataset/
    dummy_data/
    __init__.py
    my_dataset.py
    my_dataset_test.py
    make_dummy_data.py
    checksums.tsv
```

Instructions:

```
python3 -m tensorflow_datasets.scripts.cleanup.refactor_dataset_as_folder
```

"""

import dataclasses
import shutil
from typing import Type

from absl import app
from absl import flags
import tensorflow_datasets as tfds

flags.DEFINE_string('datasets', None, 'Datasets to convert')

FLAGS = flags.FLAGS

TFDS_PATH = tfds.core.utils.tfds_write_path()


@dataclasses.dataclass(frozen=True)
class BuilderCodeInfo:
  """Wrapper around `DatasetBuilder` code metadata.

  Attributes:
    file: Path to dataset file (e.g. `.../image/mnist.py`)
    dir: Path to dataset directory (e.g. `.../image/`)
    dst: New dataset directory `.../image/mnist/`
    name: Dataset name, snake_case (`my_dataset`)
    cls_name: Dataset name, CamelCase (`MyDataset`)
    type: Dataset type (e.g. `image`, `text`, `object_detection`)
  """

  file: tfds.core.Path
  dir: tfds.core.Path
  dst: tfds.core.Path
  name: str
  cls_name: str
  type: str

  @classmethod
  def from_builder_cls(
      cls, builder_cls: Type[tfds.core.DatasetBuilder]
  ) -> 'BuilderCodeInfo':
    path = tfds.core.utils.to_write_path(builder_cls.code_path)
    return cls(
        file=path,
        dir=path.parent,
        dst=path.parent / builder_cls.name,
        name=builder_cls.name,
        cls_name=builder_cls.__name__,
        type=path.parent.name,
    )


# Util functions


def _rename_dir(
    src: tfds.core.Path,
    dst: tfds.core.Path,
) -> None:
  """Equivalent of `src.rename(dst)`."""
  # src.rename(dst) creates `Invalid cross-device link` on some remote file
  # systems, so uses manual operation instead.
  # Note, this is not atomic.
  shutil.copytree(src, dst)
  shutil.rmtree(src)


# Rename functions


def _add_init_file(code_info: BuilderCodeInfo) -> None:
  """Creates the `my_dataset/__init__.py` file."""
  init_path = code_info.dst / '__init__.py'
  init_path.write_text('\n')


def _mv_fake_data_dir(code_info: BuilderCodeInfo) -> None:
  """Move the fake data directory."""
  src_fake_dir_path = (
      TFDS_PATH / 'testing/test_data/fake_examples' / code_info.name
  )
  dst_fake_dir_path = code_info.dst / 'dummy_data'
  _rename_dir(src_fake_dir_path, dst_fake_dir_path)


def _mv_code(code_info: BuilderCodeInfo) -> None:
  """Move the `my_dataset.py` file."""
  code_info.file.rename(code_info.dst / f'{code_info.name}.py')


def _mv_code_test(code_info: BuilderCodeInfo) -> None:
  """Move the `my_dataset_test.py` file."""
  # Try to infer the test filename.
  test_file = code_info.dir / f'{code_info.file.stem}_test.py'
  assert test_file.exists(), f'{test_file} not found.'
  test_file.rename(code_info.dst / f'{code_info.name}_test.py')


def _mv_checksums(code_info: BuilderCodeInfo) -> None:
  """Move the `checksums.tsv` file."""
  src_checksums_path = TFDS_PATH / 'url_checksums' / f'{code_info.name}.txt'
  # Not all datasets have checksums (e.g. manual datasets)
  if src_checksums_path.exists():
    src_checksums_path.rename(code_info.dst / 'checksums.tsv')


def _mv_create_fake_data(code_info: BuilderCodeInfo) -> None:
  """Move the `fake_data` generation script file."""
  create_fake_data_file = (
      TFDS_PATH / 'testing/fake_data_generation' / f'{code_info.name}.py'
  )
  if create_fake_data_file.exists():
    create_fake_data_file.rename(code_info.dst / 'make_dummy_data.py')


# Main functions


def refactor_dataset(ds_name: str) -> None:
  """Refactor a single dataset."""
  code_info = BuilderCodeInfo.from_builder_cls(tfds.builder_cls(ds_name))

  print(f'Refactoring {code_info.name} in {code_info.dst}')

  # Eventually cleanup previous refactoring.
  if code_info.dst.exists():
    print(f'Cleanup existing {code_info.dst}')
    shutil.rmtree(code_info.dst)
  code_info.dst.mkdir()

  # Copy all files and folders
  _add_init_file(code_info)
  _mv_code(code_info)
  _mv_code_test(code_info)
  _mv_checksums(code_info)
  _mv_fake_data_dir(code_info)
  _mv_create_fake_data(code_info)

  # TODO(tfds):
  # * Add BUILD files
  # * Update `<type>/__init__.py` import (including builder config class)
  # * Support util scripts
  # * How to support shared data (e.g. imagenet labels shared between classes ?)


def refactor_datasets() -> None:
  """Refactoring all dataset into one folder."""
  for ds_name in FLAGS.datasets.split(',') or tfds.list_builders(
      with_community_datasets=False
  ):
    refactor_dataset(ds_name)


def main(_):
  refactor_datasets()


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the GLUE data set.

We have an individual test for each config so that we can use sharding to
prevent the test from timing out.
"""

from tensorflow_datasets import testing
from tensorflow_datasets.text import glue


class GlueColaTest(testing.DatasetBuilderTestCase):
  BUILDER_CONFIG_NAMES_TO_TEST = ["cola"]
  DATASET_CLASS = glue.Glue
  SPLITS = {
      "train": 3,
      "validation": 2,
      "test": 1,
  }


class GlueSst2Test(testing.DatasetBuilderTestCase):
  BUILDER_CONFIG_NAMES_TO_TEST = ["sst2"]
  DATASET_CLASS = glue.Glue
  SPLITS = {
      "train": 3,
      "validation": 2,
      "test": 1,
  }


class GlueQqpTest(testing.DatasetBuilderTestCase):
  BUILDER_CONFIG_NAMES_TO_TEST = ["qqp"]
  DATASET_CLASS = glue.Glue
  SPLITS = {
      "train": 3,
      "validation": 2,
      "test": 1,
  }


class GlueStsbTest(testing.DatasetBuilderTestCase):
  BUILDER_CONFIG_NAMES_TO_TEST = ["stsb"]
  DATASET_CLASS = glue.Glue
  SPLITS = {
      "train": 3,
      "validation": 2,
      "test": 1,
  }


class GlueMnliTest(testing.DatasetBuilderTestCase):
  BUILDER_CONFIG_NAMES_TO_TEST = ["mnli"]
  DATASET_CLASS = glue.Glue
  SPLITS = {
      "train": 3,
      "validation_matched": 2,
      "validation_mismatched": 2,
      "test_matched": 1,
      "test_mismatched": 1,
  }


class GlueQnliTest(testing.DatasetBuilderTestCase):
  BUILDER_CONFIG_NAMES_TO_TEST = ["qnli"]
  DATASET_CLASS = glue.Glue
  SPLITS = {
      "train": 3,
      "validation": 2,
      "test": 1,
  }


class GlueRteTest(testing.DatasetBuilderTestCase):
  BUILDER_CONFIG_NAMES_TO_TEST = ["rte"]
  DATASET_CLASS = glue.Glue
  SPLITS = {
      "train": 3,
      "validation": 2,
      "test": 1,
  }


class GlueWnliTest(testing.DatasetBuilderTestCase):
  BUILDER_CONFIG_NAMES_TO_TEST = ["wnli"]
  DATASET_CLASS = glue.Glue
  SPLITS = {
      "train": 3,
      "validation": 2,
      "test": 1,
  }


class GlueMrpcTest(testing.DatasetBuilderTestCase):
  BUILDER_CONFIG_NAMES_TO_TEST = ["mrpc"]
  DATASET_CLASS = glue.Glue
  DL_EXTRACT_RESULT = {
      "train": "MRPC/msr_paraphrase_train.txt",
      "test": "MRPC/msr_paraphrase_test.txt",
      "dev_ids": "MRPC/mrpc_dev_ids.tsv",
  }
  SPLITS = {
      "train": 10,
      "validation": 8,
      "test": 15,
  }


class GlueAxTest(testing.DatasetBuilderTestCase):
  BUILDER_CONFIG_NAMES_TO_TEST = ["ax"]
  DATASET_CLASS = glue.Glue
  DL_EXTRACT_RESULT = "AX/ax.tsv"
  SPLITS = {
      "test": 3,
  }


if __name__ == "__main__":
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""dart dataset."""

from __future__ import annotations

import json
import os

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds

_CITATION = """
@article{radev2020dart,
  title={DART: Open-Domain Structured Data Record to Text Generation},
  author={Dragomir Radev and Rui Zhang and Amrit Rau and Abhinand Sivaprasad and Chiachun Hsieh and Nazneen Fatema Rajani and Xiangru Tang and Aadit Vyas and Neha Verma and Pranav Krishna and Yangxiaokang Liu and Nadia Irwanto and Jessica Pan and Faiaz Rahman and Ahmad Zaidi and Murori Mutuma and Yasin Tarabar and Ankit Gupta and Tao Yu and Yi Chern Tan and Xi Victoria Lin and Caiming Xiong and Richard Socher},
  journal={arXiv preprint arXiv:2007.02871},
  year={2020}
"""

_DESCRIPTION = """
DART (DAta Record to Text generation) contains RDF entity-relation annotated
with sentence descriptions that cover all facts in the triple set. DART was
constructed using existing datasets such as: WikiTableQuestions, WikiSQL, WebNLG
and Cleaned E2E. The tables from WikiTableQuestions and WikiSQL were transformed
to subject-predicate-object triples, and its text annotations were mainly
collected from MTurk. The meaningful representations in E2E were also
transformed to triples and its descriptions were used, some that couldn't be
transformed were dropped.

The dataset splits of E2E and WebNLG are kept, and for the WikiTableQuestions
and WikiSQL the Jaccard similarity is used to keep similar tables in the same
set (train/dev/tes).

This dataset is constructed following a standarized table format.
"""

_URL = 'https://github.com/Yale-LILY/dart/archive/master.zip'


class Dart(tfds.core.GeneratorBasedBuilder):
  """DAta Record to Text Generation."""

  VERSION = tfds.core.Version('0.1.0')

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        # This is the description that will appear on the datasets page.
        description=_DESCRIPTION,
        # tfds.features.FeatureConnectors
        features=tfds.features.FeaturesDict({
            'input_text': {
                'table': tfds.features.Sequence({  # Each row will be one triple fact.
                    # we'll only have subject/predicate/object headers
                    'column_header': np.str_,
                    'row_number': np.int16,
                    'content': np.str_,
                }),
            },
            'target_text': np.str_,
        }),
        supervised_keys=('input_text', 'target_text'),
        # Homepage of the dataset for documentation
        homepage='https://github.com/Yale-LILY/dart',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    extracted_path = os.path.join(
        dl_manager.download_and_extract(_URL), 'dart-master', 'data', 'v1.1.1'
    )
    return {
        tfds.Split.TRAIN: self._generate_examples(
            json_file=os.path.join(
                extracted_path, 'dart-v1.1.1-full-train.json'
            )
        ),
        tfds.Split.VALIDATION: self._generate_examples(
            json_file=os.path.join(extracted_path, 'dart-v1.1.1-full-dev.json')
        ),
        tfds.Split.TEST: self._generate_examples(
            json_file=os.path.join(extracted_path, 'dart-v1.1.1-full-test.json')
        ),
    }

  def _generate_examples(self, json_file):
    """Yields examples."""
    with epath.Path(json_file).open() as f:
      data = json.load(f)
      for entry_count, entry in enumerate(data):
        table = []
        for i, triple_set in enumerate(entry['tripleset']):
          for header, content in zip(
              ['subject', 'predicate', 'object'], triple_set
          ):
            table.append({
                'column_header': header,
                'row_number': i,
                'content': content,
            })
        for annotation_count, annotation in enumerate(entry['annotations']):
          yield '{}_{}'.format(entry_count, annotation_count), {
              'input_text': {
                  'table': table,
              },
              'target_text': annotation['text'],
          }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.utils.tqdm_utils."""

from tensorflow_datasets import testing
from tensorflow_datasets.core import dataset_utils
from tensorflow_datasets.core.utils import tqdm_utils


class TqdmUtilsTest(testing.TestCase):

  def test_disable_tqdm(self):
    tqdm_utils.disable_progress_bar()

    with testing.tmp_dir(self.get_temp_dir()) as tmp_dir:
      builder = testing.DummyMnist(data_dir=tmp_dir)
      builder.download_and_prepare()

      # Check the data has been generated
      train_ds, test_ds = builder.as_dataset(split=['train', 'test'])
      train_ds, test_ds = dataset_utils.as_numpy((train_ds, test_ds))
      self.assertEqual(20, len(list(train_ds)))
      self.assertEqual(20, len(list(test_ds)))

  def test_disable_tqdm2(self):
    tqdm_utils.display_progress_bar(enable=False)

    with testing.tmp_dir(self.get_temp_dir()) as tmp_dir:
      builder = testing.DummyMnist(data_dir=tmp_dir)
      builder.download_and_prepare()

      # Check the data has been generated
      train_ds, test_ds = builder.as_dataset(split=['train', 'test'])
      train_ds, test_ds = dataset_utils.as_numpy((train_ds, test_ds))
      self.assertEqual(20, len(list(train_ds)))
      self.assertEqual(20, len(list(test_ds)))


if __name__ == '__main__':
  testing.test_main()
